from dataclasses import dataclass
from typing import (
    Optional,
    Sequence,
)

from pcs.common.const import PcmkAction
from pcs.common.interface.dto import DataTransferObject
from pcs.common.pacemaker.types import CibResourceSetOrderType

from .set import CibResourceSetDto


@dataclass(frozen=True)
class CibConstraintOrderAttributesDto(DataTransferObject):
    constraint_id: str
    symmetrical: Optional[bool]
    require_all: Optional[bool]
    score: Optional[str]
    kind: Optional[CibResourceSetOrderType]


@dataclass(frozen=True)
class CibConstraintOrderDto(DataTransferObject):
    first_resource_id: str
    then_resource_id: str
    first_action: Optional[PcmkAction]
    then_action: Optional[PcmkAction]
    first_resource_instance: Optional[int]
    then_resource_instance: Optional[int]
    attributes: CibConstraintOrderAttributesDto


@dataclass(frozen=True)
class CibConstraintOrderSetDto(DataTransferObject):
    resource_sets: Sequence[CibResourceSetDto]
    attributes: CibConstraintOrderAttributesDto



import json
from typing import Any

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import (
    Argv,
    InputModifiers,
    KeyValueParser,
    group_by_keywords,
)


def alert_add(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file (in lib wrapper)
    """
    modifiers.ensure_only_supported("-f")
    if not argv:
        raise CmdLineInputError()

    sections = group_by_keywords(
        argv, {"options", "meta"}, implicit_first_keyword="main"
    )
    parser = KeyValueParser(sections.get_args_flat("main"))
    parser.check_allowed_keys(["id", "description", "path"])
    main_args = parser.get_unique()

    lib.alert.create_alert(
        main_args.get("id", None),
        main_args.get("path", None),
        KeyValueParser(sections.get_args_flat("options")).get_unique(),
        KeyValueParser(sections.get_args_flat("meta")).get_unique(),
        main_args.get("description", None),
    )


def alert_update(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file (in lib wrapper)
    """
    modifiers.ensure_only_supported("-f")
    if not argv:
        raise CmdLineInputError()

    alert_id = argv[0]

    sections = group_by_keywords(
        argv[1:], {"options", "meta"}, implicit_first_keyword="main"
    )
    parser = KeyValueParser(sections.get_args_flat("main"))
    parser.check_allowed_keys(["description", "path"])
    main_args = parser.get_unique()

    lib.alert.update_alert(
        alert_id,
        main_args.get("path", None),
        KeyValueParser(sections.get_args_flat("options")).get_unique(),
        KeyValueParser(sections.get_args_flat("meta")).get_unique(),
        main_args.get("description", None),
    )


def alert_remove(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file (in lib wrapper)
    """
    modifiers.ensure_only_supported("-f")
    if not argv:
        raise CmdLineInputError()

    lib.alert.remove_alert(argv)


def recipient_add(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file (in lib wrapper)
      * --force - allows not unique recipient values
    """
    modifiers.ensure_only_supported("-f", "--force")
    if len(argv) < 2:
        raise CmdLineInputError()

    alert_id = argv[0]

    sections = group_by_keywords(
        argv[1:], {"options", "meta"}, implicit_first_keyword="main"
    )
    parser = KeyValueParser(sections.get_args_flat("main"))
    parser.check_allowed_keys(["description", "id", "value"])
    main_args = parser.get_unique()

    lib.alert.add_recipient(
        alert_id,
        main_args.get("value", None),
        KeyValueParser(sections.get_args_flat("options")).get_unique(),
        KeyValueParser(sections.get_args_flat("meta")).get_unique(),
        recipient_id=main_args.get("id", None),
        description=main_args.get("description", None),
        allow_same_value=modifiers.get("--force"),
    )


def recipient_update(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file (in lib wrapper)
      * --force - allows not unique recipient values
    """
    modifiers.ensure_only_supported("-f", "--force")
    if not argv:
        raise CmdLineInputError()

    recipient_id = argv[0]

    sections = group_by_keywords(
        argv[1:], {"options", "meta"}, implicit_first_keyword="main"
    )
    parser = KeyValueParser(sections.get_args_flat("main"))
    parser.check_allowed_keys(["description", "value"])
    main_args = parser.get_unique()

    lib.alert.update_recipient(
        recipient_id,
        KeyValueParser(sections.get_args_flat("options")).get_unique(),
        KeyValueParser(sections.get_args_flat("meta")).get_unique(),
        recipient_value=main_args.get("value", None),
        description=main_args.get("description", None),
        allow_same_value=modifiers.get("--force"),
    )


def recipient_remove(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file (in lib wrapper)
    """
    modifiers.ensure_only_supported("-f")
    if not argv:
        raise CmdLineInputError()

    lib.alert.remove_recipient(argv)


def print_alerts_in_json(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    This is used only by pcsd, will be removed in new architecture

    Options:
      * -f - CIB file (in lib wrapper)
    """
    modifiers.ensure_only_supported("-f")
    if argv:
        raise CmdLineInputError()

    print(json.dumps(lib.alert.get_all_alerts()))



from typing import Mapping

from lxml.etree import _Element

from pcs.common import reports
from pcs.common.services.interfaces import ServiceManagerInterface
from pcs.common.tools import timeout_to_seconds
from pcs.common.types import StringSequence
from pcs.lib import sbd, validate
from pcs.lib.cib import nvpair_multi
from pcs.lib.cib.tools import (
    IdProvider,
    get_crm_config,
    get_pacemaker_version_by_which_cib_was_validated,
)
from pcs.lib.errors import LibraryError
from pcs.lib.external import CommandRunner
from pcs.lib.pacemaker.values import is_false
from pcs.lib.resource_agent import ResourceAgentFacade

READONLY_CLUSTER_PROPERTY_LIST = [
    "cluster-infrastructure",
    "cluster-name",
    "dc-version",
    "have-watchdog",
    "last-lrm-refresh",
]
_DEFAULT_CLUSTER_PROPERTY_SET_ID = "cib-bootstrap-options"
_STONITH_WATCHDOG_TIMEOUT_PROPERTIES = [
    "stonith-watchdog-timeout",
    "fencing-watchdog-timeout",
]


def _validate_stonith_watchdog_timeout_property(
    service_manager: ServiceManagerInterface,
    value: str,
    force: bool = False,
) -> reports.ReportItemList:
    report_list: reports.ReportItemList = []
    original_value = value
    # if value is not empty, try to convert time interval string
    if value:
        seconds = timeout_to_seconds(value)
        if seconds is None:
            # returns empty list because this should be reported by
            # ValueTimeInterval validator
            return report_list
        value = str(seconds)
    if sbd.is_sbd_enabled(service_manager):
        report_list.extend(
            sbd.validate_stonith_watchdog_timeout(
                validate.ValuePair(original_value, value), force
            )
        )
    elif value not in ["", "0"]:
        report_list.append(
            reports.ReportItem.error(
                reports.messages.StonithWatchdogTimeoutCannotBeSet(
                    reports.const.SBD_NOT_SET_UP
                ),
            )
        )
    return report_list


def _validate_not_disabling_fencing(
    to_be_set_properties: Mapping[str, str],
) -> reports.ReportItemList:
    problematic_properties_setting = {
        key: to_be_set_properties[key]
        for key in ["stonith-enabled", "fencing-enabled"]
        if key in to_be_set_properties and is_false(to_be_set_properties[key])
    }

    if problematic_properties_setting:
        return [
            reports.ReportItem.warning(
                reports.messages.NoStonithMeansWouldBeLeftDueToProperties(
                    problematic_properties_setting
                ),
            )
        ]

    return []


def validate_set_cluster_properties(  # noqa: PLR0912
    runner: CommandRunner,
    cluster_properties_facade: ResourceAgentFacade,
    properties_set_id: str,
    configured_properties: StringSequence,
    new_properties: Mapping[str, str],
    service_manager: ServiceManagerInterface,
    force: bool = False,
) -> reports.ReportItemList:
    """
    Validate that cluster properties and their values can be set.

    cluster_properties_facade -- facade for cluster properties metadata
    properties_set_id -- id of the properties set to be updated
    configured_properties -- names of currently configured cluster properties
    new_properties -- dictionary of properties and their values to be set
    service_manager -- manager for system daemon services
    force -- if True, produce warnings instead of errors
    """
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-locals
    possible_properties_dict = {
        parameter.name: parameter
        for parameter in cluster_properties_facade.metadata.parameters
        if parameter.name not in READONLY_CLUSTER_PROPERTY_LIST
    }
    severity = reports.get_severity(reports.codes.FORCE, force)

    to_be_set_properties = {}
    to_be_removed_properties = []
    for name, value in new_properties.items():
        if value != "":
            to_be_set_properties[name] = value
        else:
            to_be_removed_properties.append(name)

    report_list = []

    report_list.extend(
        validate.ValidatorAll(
            cluster_properties_facade.get_validators_deprecated_parameters()
        ).validate(new_properties)
    )
    report_list.extend(
        validate.validate_set_unset_items(
            to_be_set_properties.keys(),
            to_be_removed_properties,
            configured_properties,
            reports.const.ADD_REMOVE_CONTAINER_TYPE_PROPERTY_SET,
            reports.const.ADD_REMOVE_ITEM_TYPE_PROPERTY,
            properties_set_id,
            severity=severity,
        )
    )
    report_list.extend(
        validate.ValidatorAll(
            cluster_properties_facade.get_validators_allowed_parameters(
                force=force,
                banned_parameter_list=READONLY_CLUSTER_PROPERTY_LIST,
            )
        ).validate(
            # Allow removing properties unknown to pacemaker while preventing
            # setting them. Prevent removing read-only properties.
            {
                name: value
                for name, value in new_properties.items()
                if value != "" or name in READONLY_CLUSTER_PROPERTY_LIST
            }
        )
    )

    # Validators are based on pacemaker validation:
    #   * https://github.com/ClusterLabs/pacemaker/blob/main/lib/common/options.c
    # There are no pacemaker validators for these types:
    #   * string - just arbitrary strings
    #   * version - used only by pacemaker generated option 'dc-version'
    #   * epoch_time - type is not used at all
    validators: list[validate.ValidatorInterface] = []
    for property_name in to_be_set_properties:
        if property_name not in possible_properties_dict:
            # unknown properties are reported by NamesIn validator
            continue
        property_metadata = possible_properties_dict[property_name]
        if property_metadata.type == "boolean":
            validators.append(
                validate.ValuePcmkBoolean(
                    property_metadata.name, severity=severity
                )
            )
        elif property_metadata.type == "duration":
            validators.append(
                validate.ValueTimeIntervalOrDuration(
                    runner, property_metadata.name, severity=severity
                )
            )
        elif property_metadata.type in ["integer", "score"]:
            validators.append(
                validate.ValuePcmkInteger(
                    property_metadata.name, severity=severity
                )
            )
        elif property_metadata.type == "nonnegative_integer":
            validators.append(
                validate.ValuePcmkPositiveInteger(
                    property_metadata.name, severity=severity
                )
            )
        elif property_metadata.type == "percentage":
            validators.append(
                validate.ValuePcmkPercentage(
                    property_metadata.name, severity=severity
                )
            )
        elif property_metadata.type == "port":
            validators.append(
                validate.ValuePortNumber(
                    property_metadata.name, severity=severity
                )
            )
        elif property_metadata.type == "select":
            validators.append(
                validate.ValueIn(
                    property_metadata.name,
                    property_metadata.enum_values or [],
                    severity=severity,
                )
            )
        elif property_metadata.type in ["time", "timeout"]:
            # make stonith-watchdog-timeout value not forcable
            if property_metadata.name in _STONITH_WATCHDOG_TIMEOUT_PROPERTIES:
                validators.append(
                    validate.ValueTimeInterval(
                        property_metadata.name,
                        severity=reports.ReportItemSeverity.error(),
                    )
                )
            elif property_metadata.type == "timeout":
                validators.append(
                    validate.ValueTimeInterval(
                        property_metadata.name, severity=severity
                    )
                )
            else:  # time
                validators.append(
                    validate.ValueTimeIntervalOrDuration(
                        runner, property_metadata.name, severity=severity
                    )
                )

    report_list.extend(
        validate.ValidatorAll(validators).validate(to_be_set_properties)
    )

    # Only validate SWT if it is being set, or if it is being removed and it
    # actually exists in the current configuration.
    for prop_name in _STONITH_WATCHDOG_TIMEOUT_PROPERTIES:
        if prop_name in new_properties and (
            new_properties[prop_name] or prop_name in configured_properties
        ):
            report_list.extend(
                _validate_stonith_watchdog_timeout_property(
                    service_manager,
                    new_properties[prop_name],
                    force=force,
                )
            )

    report_list.extend(_validate_not_disabling_fencing(to_be_set_properties))

    return report_list


def get_cluster_property_set_element_legacy(
    cib: _Element, id_provider: IdProvider
) -> _Element:
    """
    Return the first cluster_property_set element. If the element does not
    exist, try to create cluster_property_set element with id of value
    'cib-bootstrap-options'. Raise an error in case the id is already used.

    cib -- cib tree
    id_provider -- checks id uniqueness and books ids if set
    """
    crm_config_el = get_crm_config(cib)
    property_el_list = nvpair_multi.find_nvsets(
        crm_config_el, nvpair_multi.NVSET_PROPERTY
    )
    if property_el_list:
        return property_el_list[0]
    if id_provider.book_ids(_DEFAULT_CLUSTER_PROPERTY_SET_ID):
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.CannotCreateDefaultClusterPropertySet(
                    _DEFAULT_CLUSTER_PROPERTY_SET_ID
                )
            )
        )
    return nvpair_multi.nvset_append_new(
        crm_config_el,
        id_provider,
        get_pacemaker_version_by_which_cib_was_validated(cib),
        nvpair_multi.NVSET_PROPERTY,
        {},
        {"id": _DEFAULT_CLUSTER_PROPERTY_SET_ID},
        nvset_rule=None,
    )



from lxml.etree import _Element

from pcs.common import reports
from pcs.common.types import StringCollection
from pcs.lib.cib.resource.common import is_disabled as is_resource_disabled
from pcs.lib.cib.resource.stonith import get_all_node_isolating_resources
from pcs.lib.communication.sbd import GetSbdStatus
from pcs.lib.communication.tools import run as run_communication
from pcs.lib.env import LibraryEnvironment
from pcs.lib.node import get_existing_nodes_names


def ensure_some_stonith_remains(
    env: LibraryEnvironment,
    resources_el: _Element,
    stonith_resources_to_ignore: StringCollection,
    sbd_being_disabled: bool,
    force_flags: reports.types.ForceFlags,
) -> reports.ReportItemList:
    """
    Error out when no sbd or enabled stonith would be left after a config change

    resource_el -- cib element holding resources
    stonith_resources_to_ignore -- ids of stonith being removed, disabled, etc.
    sbd_being_disabled -- ignore working sbd as it is being disabled
    force_flags -- use to emit a warning instead of an error
    """
    # Checking whether sbd is enabled requires communicating with other cluster
    # nodes, which may bring additional issues, like nodes not being
    # accessible. To reduce clutter in reports, check for the sbd being enabled
    # only when necessary.

    if not stonith_resources_to_ignore and not sbd_being_disabled:
        # No stonith resources are being removed or disabled and SBD is not
        # being disabled either. There is no change in cluster fencing
        # capabilities and therefore nothing to report.
        return []

    current_stonith = [
        stonith_el
        for stonith_el in get_all_node_isolating_resources(resources_el)
        # If any nvset disables the resource, even with a rule to limit it to
        # specific time, then the resource wouldn't be able to fence all the
        # time and should be considered disabled.
        # However, pcs currently supports only one nvset for meta attributes,
        # so we only check that to be consistent. Checking all nvsets could
        # lead to a situation not resolvable by pcs, as pcs doesn't allow to
        # change other nvsets than the first one.
        # Technically, stonith resources can be disabled by their parent clones
        # or groups. However, pcs doesn't allow putting stonith to groups and
        # clones, so we don't check that.
        # The check is not perfect, but it is a reasonable effort, considering
        # that multiple nvsets are not supported for meta attributes by pcs now.
        # It can be improved when a need for it raises.
        if not is_resource_disabled(stonith_el)
    ]
    stonith_left = [
        stonith_el
        for stonith_el in current_stonith
        if stonith_el.attrib["id"] not in stonith_resources_to_ignore
    ]

    if stonith_left:
        # Working stonith devices will be present. No need to check for SBD.
        return []

    # No stonith in the cluster, need to check SBD.
    current_sbd_active = _is_sbd_active_on_any_node(env)
    sbd_left_active = current_sbd_active and not sbd_being_disabled

    if sbd_left_active:
        # SBD will be left enabled.
        return []

    if not current_stonith and not current_sbd_active:
        # Now we know that no enabled stonith will be left in the cluster and
        # sbd will also be disabled. However, if that already was the case, we
        # don't produce an error -> the cluster already cannot fence, saying
        # that it is a result of the current change would not be true.
        return []

    return [
        reports.ReportItem(
            reports.get_severity(
                reports.codes.FORCE, reports.codes.FORCE in force_flags
            ),
            reports.messages.NoStonithMeansWouldBeLeft(),
        )
    ]


def _is_sbd_active_on_any_node(env: LibraryEnvironment) -> bool:
    # SBD can be enabled only partially in the cluster. Even when that is the
    # case, we warn the user when disabling it. For example, SBD can be enabled
    # for full stack nodes and disabled for remote / guest nodes.
    if not env.is_cib_live:
        # We cannot tell whether sbd is enabled or not, as we do not have
        # access to a cluster. Expect sbd is not enabled.
        return False

    # Do not return errors. The check should not prevent deleting a resource
    # just because a node in a cluster is temporarily unavailable. We do our
    # best to figure out sbd status.
    node_list, get_nodes_report_list = get_existing_nodes_names(
        env.get_corosync_conf()
    )
    env.report_processor.report_list(get_nodes_report_list)
    if not node_list:
        env.report_processor.report(
            reports.ReportItem.warning(
                reports.messages.CorosyncConfigNoNodesDefined()
            )
        )
        return False

    com_cmd = GetSbdStatus(env.report_processor)
    com_cmd.set_targets(
        env.get_node_target_factory().get_target_list(
            node_list, skip_non_existing=True
        )
    )
    response_per_node = run_communication(env.get_node_communicator(), com_cmd)
    for response in response_per_node:
        # Values can be either True (== sbd is enabled), or False (== sbd is
        # disabled), or None (== unknown, not connected).
        # We do not want to block removing resources just because we were
        # temporarily unable to connect to a node.
        # If sbd is enabled and not running, then the cluster won't have any
        # fencing after removing all stonith resources. If sbd is not enabled
        # and running, then the cluster won't have any fencing after removing
        # all stonith resources and rebooting nodes once. So we need both
        # enabled and running to be true to consider sbd as active.
        if response["status"]["enabled"] and response["status"]["running"]:
            return True
    return False



from unittest import TestCase, mock

from pcs import settings
from pcs.common import reports
from pcs.lib.commands import cluster_property

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import get_test_resource as rc

from .crm_attribute_mixins import CrmAttributeLoadMetadataMixin
from .fixtures import fixture_crm_config_properties


class StonithWatchdogTimeoutTestCase(CrmAttributeLoadMetadataMixin):
    sbd_enabled = None

    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.config.runner.cib.load(
            crm_config=fixture_crm_config_properties(
                [("cib-bootstrap-options", {"stonith-watchdog-timeout": "10"})]
            )
        )
        self.load_fake_agent_metadata()
        self.config.services.is_enabled("sbd", return_value=self.sbd_enabled)

    def _set_success(self, options_dict):
        self.config.env.push_cib(
            crm_config=fixture_crm_config_properties(
                [("cib-bootstrap-options", options_dict)]
            )
        )
        cluster_property.set_properties(
            self.env_assist.get_env(), options_dict, []
        )
        self.env_assist.assert_reports([])

    def _set_invalid_value(self, forced=False):
        self.config.remove("services.is_enabled")
        self.env_assist.assert_raise_library_error(
            lambda: cluster_property.set_properties(
                self.env_assist.get_env(),
                {"stonith-watchdog-timeout": "15x"},
                [] if not forced else [reports.codes.FORCE],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="stonith-watchdog-timeout",
                    option_value="15x",
                    allowed_values="time interval (e.g. 1, 2s, 3m, 4h, ...)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ]
        )

    def test_set_invalid_value(self):
        self._set_invalid_value(forced=False)

    def test_set_invalid_value_forced(self):
        self._set_invalid_value(forced=True)


class CannotSetNoemptyProperty(StonithWatchdogTimeoutTestCase):
    not_forced_reports = []

    def test_set_empty(self):
        self._set_success({"stonith-watchdog-timeout": ""})

    def test_set_zero(self):
        self._set_success({"stonith-watchdog-timeout": "0"})

    def test_set_to_zero_time_suffix(self):
        self._set_success({"stonith-watchdog-timeout": "0min"})

    def test_set_not_zero_or_empty(self):
        self.env_assist.assert_raise_library_error(
            lambda: cluster_property.set_properties(
                self.env_assist.get_env(),
                {"stonith-watchdog-timeout": "20"},
                [],
            )
        )
        self.env_assist.assert_reports(self.not_forced_reports)


@mock.patch.object(
    settings,
    "pacemaker_api_result_schema",
    rc("pcmk_rng/api/api-result.rng"),
)
class TestSetStonithWatchdogTimeoutSBDDisabled(
    CannotSetNoemptyProperty, TestCase
):
    sbd_enabled = False
    not_forced_reports = [
        fixture.error(
            reports.codes.STONITH_WATCHDOG_TIMEOUT_CANNOT_BE_SET,
            reason="sbd_not_set_up",
        ),
    ]

    def test_set_not_zero_or_empty_forced(self):
        self.env_assist.assert_raise_library_error(
            lambda: cluster_property.set_properties(
                self.env_assist.get_env(),
                {"stonith-watchdog-timeout": "20"},
                [reports.codes.FORCE],
            )
        )
        self.env_assist.assert_reports(self.not_forced_reports)


@mock.patch("pcs.lib.sbd._get_local_sbd_watchdog_timeout", lambda: 10)
@mock.patch("pcs.lib.sbd.get_local_sbd_device_list", list)
class TestSetStonithWatchdogTimeoutSBDEnabledDevicesUnspecified(
    StonithWatchdogTimeoutTestCase, TestCase
):
    sbd_enabled = True

    def test_set_empty(self):
        self.env_assist.assert_raise_library_error(
            lambda: cluster_property.set_properties(
                self.env_assist.get_env(), {"stonith-watchdog-timeout": ""}, []
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.STONITH_WATCHDOG_TIMEOUT_CANNOT_BE_UNSET,
                    force_code=reports.codes.FORCE,
                    reason="sbd_set_up_without_devices",
                )
            ]
        )

    def test_set_empty_forced(self):
        self.config.env.push_cib(
            crm_config=fixture_crm_config_properties(
                [("cib-bootstrap-options", {})]
            )
        )
        cluster_property.set_properties(
            self.env_assist.get_env(),
            {"stonith-watchdog-timeout": ""},
            [reports.codes.FORCE],
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.STONITH_WATCHDOG_TIMEOUT_CANNOT_BE_UNSET,
                    reason="sbd_set_up_without_devices",
                )
            ]
        )

    def test_set_zero(self):
        self.env_assist.assert_raise_library_error(
            lambda: cluster_property.set_properties(
                self.env_assist.get_env(),
                {"stonith-watchdog-timeout": "0"},
                [],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.STONITH_WATCHDOG_TIMEOUT_CANNOT_BE_UNSET,
                    force_code=reports.codes.FORCE,
                    reason="sbd_set_up_without_devices",
                )
            ]
        )

    def test_set_zero_forced(self):
        self.config.env.push_cib(
            crm_config=fixture_crm_config_properties(
                [("cib-bootstrap-options", {"stonith-watchdog-timeout": "0s"})]
            )
        )
        cluster_property.set_properties(
            self.env_assist.get_env(),
            {"stonith-watchdog-timeout": "0s"},
            [reports.codes.FORCE],
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.STONITH_WATCHDOG_TIMEOUT_CANNOT_BE_UNSET,
                    reason="sbd_set_up_without_devices",
                )
            ]
        )

    def test_equal_to_timeout(self):
        self.env_assist.assert_raise_library_error(
            lambda: cluster_property.set_properties(
                self.env_assist.get_env(),
                {"stonith-watchdog-timeout": "10"},
                [],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.STONITH_WATCHDOG_TIMEOUT_TOO_SMALL,
                    force_code=reports.codes.FORCE,
                    cluster_sbd_watchdog_timeout=10,
                    entered_watchdog_timeout="10",
                )
            ]
        )

    def test_too_small(self):
        self.env_assist.assert_raise_library_error(
            lambda: cluster_property.set_properties(
                self.env_assist.get_env(),
                {"stonith-watchdog-timeout": "9s"},
                [],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.STONITH_WATCHDOG_TIMEOUT_TOO_SMALL,
                    force_code=reports.codes.FORCE,
                    cluster_sbd_watchdog_timeout=10,
                    entered_watchdog_timeout="9s",
                )
            ]
        )

    def test_too_small_forced(self):
        self.config.env.push_cib(
            crm_config=fixture_crm_config_properties(
                [("cib-bootstrap-options", {"stonith-watchdog-timeout": "9s"})]
            )
        )
        cluster_property.set_properties(
            self.env_assist.get_env(),
            {"stonith-watchdog-timeout": "9s"},
            [reports.codes.FORCE],
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.STONITH_WATCHDOG_TIMEOUT_TOO_SMALL,
                    cluster_sbd_watchdog_timeout=10,
                    entered_watchdog_timeout="9s",
                )
            ]
        )

    def test_more_than_timeout(self):
        self._set_success({"stonith-watchdog-timeout": "11s"})


@mock.patch.object(
    settings,
    "pacemaker_api_result_schema",
    rc("pcmk_rng/api/api-result.rng"),
)
@mock.patch("pcs.lib.sbd.get_local_sbd_device_list", lambda: ["dev1", "dev2"])
class TestSetStonithWatchdogTimeoutSBDEnabledDevicesSpecified(
    CannotSetNoemptyProperty, TestCase
):
    sbd_enabled = True
    not_forced_reports = [
        fixture.error(
            reports.codes.STONITH_WATCHDOG_TIMEOUT_CANNOT_BE_SET,
            force_code=reports.codes.FORCE,
            reason="sbd_set_up_with_devices",
        )
    ]

    def test_set_not_zero_or_empty_forced(self):
        self.config.env.push_cib(
            crm_config=fixture_crm_config_properties(
                [("cib-bootstrap-options", {"stonith-watchdog-timeout": "20"})]
            )
        )
        cluster_property.set_properties(
            self.env_assist.get_env(),
            {"stonith-watchdog-timeout": "20"},
            [reports.codes.FORCE],
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.STONITH_WATCHDOG_TIMEOUT_CANNOT_BE_SET,
                    reason="sbd_set_up_with_devices",
                )
            ]
        )



from typing import Mapping, Optional

from lxml import etree
from lxml.etree import _Element

from pcs.common import (
    const,
    reports,
)
from pcs.common.pacemaker.constraint import (
    CibConstraintLocationAttributesDto,
    CibConstraintLocationDto,
    CibConstraintLocationSetDto,
)
from pcs.common.pacemaker.role import (
    get_value_for_cib as get_role_value_for_cib,
)
from pcs.common.pacemaker.types import CibResourceDiscovery
from pcs.lib import validate
from pcs.lib.cib import rule
from pcs.lib.cib.const import TAG_CONSTRAINT_LOCATION as TAG
from pcs.lib.cib.const import TAG_RULE
from pcs.lib.cib.tools import (
    IdProvider,
    Version,
    role_constructor,
)
from pcs.lib.pacemaker.values import sanitize_id
from pcs.lib.tools import get_optional_value

from .common import (
    DuplicatesChecker,
    is_set_constraint,
    validate_constrainable_elements,
)
from .resource_set import constraint_element_to_resource_set_dto_list


def is_location_constraint(element: _Element) -> bool:
    return element.tag == TAG


def is_location_constraint_with_rule(element: _Element) -> bool:
    return (
        is_location_constraint(element)
        and element.find(f"./{TAG_RULE}") is not None
    )


def is_location_rule(element: _Element) -> bool:
    parent = element.getparent()
    return parent is not None and element.tag == TAG_RULE and parent.tag == TAG


def _element_to_attributes_dto(
    element: _Element, rule_in_effect_eval: rule.RuleInEffectEval
) -> CibConstraintLocationAttributesDto:
    return CibConstraintLocationAttributesDto(
        constraint_id=str(element.attrib["id"]),
        score=element.get("score"),
        node=element.get("node"),
        rules=[
            rule.rule_element_to_dto(rule_in_effect_eval, rule_el)
            for rule_el in element.findall(f"./{TAG_RULE}")
        ],
        lifetime=[
            rule.rule_element_to_dto(rule_in_effect_eval, rule_el)
            for rule_el in element.findall("./lifetime/rule")
        ],
        resource_discovery=get_optional_value(
            CibResourceDiscovery, element.get("resource-discovery")
        ),
    )


def _plain_constraint_el_to_dto(
    element: _Element, rule_in_effect_eval: rule.RuleInEffectEval
) -> CibConstraintLocationDto:
    return CibConstraintLocationDto(
        resource_id=element.get("rsc"),
        resource_pattern=element.get("rsc-pattern"),
        role=get_optional_value(role_constructor, element.get("role")),
        attributes=_element_to_attributes_dto(element, rule_in_effect_eval),
    )


def _set_constraint_el_to_dto(
    element: _Element, rule_in_effect_eval: rule.RuleInEffectEval
) -> CibConstraintLocationSetDto:
    return CibConstraintLocationSetDto(
        resource_sets=constraint_element_to_resource_set_dto_list(element),
        attributes=_element_to_attributes_dto(element, rule_in_effect_eval),
    )


def get_all_as_dtos(
    constraints_el: _Element, rule_in_effect_eval: rule.RuleInEffectEval
) -> tuple[list[CibConstraintLocationDto], list[CibConstraintLocationSetDto]]:
    plain_list: list[CibConstraintLocationDto] = []
    set_list: list[CibConstraintLocationSetDto] = []
    for constraint_el in constraints_el.findall(f"./{TAG}"):
        if is_set_constraint(constraint_el):
            set_list.append(
                _set_constraint_el_to_dto(constraint_el, rule_in_effect_eval)
            )
        else:
            plain_list.append(
                _plain_constraint_el_to_dto(constraint_el, rule_in_effect_eval)
            )
    return plain_list, set_list


class DuplicatesCheckerLocationRulePlain(DuplicatesChecker):
    """
    Searcher of duplicate plain location constraints with rules

    The constraint_to_check is limited to contain one rule at most. If you need
    more, split it to individual constraint elements containing one rule each
    for the purposes of the check.
    """

    def __init__(self) -> None:
        super().__init__()
        self._rule_to_str = rule.RuleToStr(normalize=True)
        self._constraint_to_check_rule: Optional[str] = None

    def _check_init(self, constraint_to_check: _Element) -> None:
        self._constraint_to_check_rule = None
        if len(constraint_to_check.findall(TAG_RULE)) != 1:
            raise RuntimeError(
                "constraint_to_check must contain exactly one rule"
            )

    def _are_duplicate(
        self,
        constraint_to_check: _Element,
        constraint_el: _Element,
    ) -> bool:
        # simple node-score constraints are not duplicate to rule constraints
        if not is_location_constraint_with_rule(constraint_el):
            return False

        # get the base constraint's rule as string
        if self._constraint_to_check_rule is None:
            rule_el = constraint_to_check.find(TAG_RULE)
            if rule_el is not None:
                self._constraint_to_check_rule = self._rule_to_str.get_str(
                    rule_el
                )

        # get the tested constraint's rules as strings
        rules = {
            self._rule_to_str.get_str(rule_el)
            for rule_el in constraint_el.iterfind(TAG_RULE)
        }

        # compare the two constraints
        return (
            constraint_to_check.get("rsc") == constraint_el.get("rsc")
            and constraint_to_check.get("rsc-pattern")
            == constraint_el.get("rsc-pattern")
            # From pacemaker explained:
            # A location constraint may contain one or more top-level rules.
            # The cluster will act as if there is a separate location
            # constraint for each rule that evaluates as true.
            and self._constraint_to_check_rule in rules
        )


class ValidateCreatePlainWithRule:
    """
    Validator for creating new constraint with a rule and appending it to CIB
    """

    def __init__(
        self,
        id_provider: IdProvider,
        rule_str: str,
        rule_options: validate.TypeOptionMap,
        constraint_options: validate.TypeOptionMap,
        constrained_el: Optional[_Element] = None,
    ):
        """
        id_provider -- elements' ids generator
        rule_str -- rule as a string, to be parsed and validated
        rule_options -- additional options for the rule
        constraint_options -- additional options for the constraint
        constrained_el -- an element for which the constraint is being created
        """
        self._id_provider = id_provider
        self._rule_str = rule_str
        self._rule_options = rule_options
        self._rule_parsed: Optional[rule.RuleRoot] = None
        self._resource_el = constrained_el
        self._constraint_options = constraint_options

    def validate(
        self, force_flags: reports.types.ForceFlags = ()
    ) -> reports.ReportItemList:
        """
        Run validation

        force_flags -- list of flags codes
        """
        force_options = reports.codes.FORCE in force_flags
        allow_in_multiinstance_resources = reports.codes.FORCE in force_flags
        report_list: reports.ReportItemList = []

        # validate resource specification
        if self._resource_el is not None:
            report_list.extend(
                validate_constrainable_elements(
                    [self._resource_el], allow_in_multiinstance_resources
                )
            )

        # validate constraint options
        validators_constraint = [
            validate.NamesIn(
                ("id", "resource-discovery"), option_type="constraint"
            ),
            # with id_provider it validates that the id is available as well
            validate.ValueId(
                "id",
                option_name_for_report="constraint id",
                id_provider=self._id_provider,
            ),
            validate.ValueIn(
                "resource-discovery",
                [
                    CibResourceDiscovery.ALWAYS,
                    CibResourceDiscovery.EXCLUSIVE,
                    CibResourceDiscovery.NEVER,
                ],
                severity=reports.item.get_severity(
                    reports.codes.FORCE, force_options
                ),
            ),
        ]
        report_list.extend(
            validate.ValidatorAll(validators_constraint).validate(
                self._constraint_options
            )
        )

        # validate rule options
        report_list.extend(self._validate_rule_options())
        # parse and validate rule
        report_list.extend(self._validate_rule())

        return report_list

    def get_parsed_rule(self) -> rule.RuleRoot:
        """
        Return validated and parsed rule for further use
        """
        if not self._rule_parsed:
            raise RuntimeError(
                "There is no valid rule. Were validator results ignored?"
            )
        return self._rule_parsed

    def _validate_rule_options(self) -> reports.ReportItemList:
        validators_rule = [
            validate.NamesIn(
                ("id", "role", "score", "score-attribute"),
                option_type="rule",
            ),
            validate.MutuallyExclusive(
                ("score", "score-attribute"),
                option_type="rule",
            ),
            # with id_provider it validates that the id is available as well
            validate.ValueId(
                "id",
                option_name_for_report="rule id",
                id_provider=self._id_provider,
            ),
            validate.ValueScore("score"),
            validate.ValueIn(
                "role",
                (const.PCMK_ROLE_PROMOTED, const.PCMK_ROLE_UNPROMOTED),
            ),
        ]
        return validate.ValidatorAll(validators_rule).validate(
            self._rule_options
        )

    def _validate_rule(self) -> reports.ReportItemList:
        report_list: reports.ReportItemList = []
        if not self._rule_str.strip():
            report_list.append(
                reports.ReportItem.error(
                    reports.messages.RuleNoExpressionSpecified()
                )
            )
        else:
            try:
                self._rule_parsed = rule.parse_rule(self._rule_str)
                report_list.extend(
                    rule.RuleValidator(
                        self._rule_parsed,
                        allow_node_attr_expr=True,
                    ).get_reports()
                )
            except rule.RuleParseError as e:
                report_list.append(
                    reports.ReportItem.error(
                        reports.messages.RuleExpressionParseError(
                            e.rule_string,
                            e.msg,
                            e.rule_line,
                            e.lineno,
                            e.colno,
                            e.pos,
                        )
                    )
                )
        return report_list


def create_plain_with_rule(
    parent_element: _Element,
    id_provider: IdProvider,
    cib_schema_version: Version,
    resource_id_type: const.ResourceIdType,
    resource_id: str,
    rule_tree: rule.RuleRoot,
    rule_options: Mapping[str, str],
    constraint_options: Mapping[str, str],
) -> _Element:
    """
    Create a location constraint with a rule for a resource

    parent_element -- where to place the constraint
    id_provider -- elements' ids generator
    cib_schema_version -- current CIB schema version
    resource_id_type -- specifies type of resource_id - plain or regexp
    resource_id -- resource ID or resource ID pattern (regexp)
    rule_tree -- parsed rule - specifies when the constraint is active
    rule_options -- additional options for the rule
    constraint_options -- additional options for the constraint
    """
    constraint_options = dict(constraint_options)  # make a modifiable copy

    # create a constraint element
    constraint_el = etree.SubElement(parent_element, TAG)

    # set constraint attributes
    if "id" not in constraint_options or not constraint_options["id"]:
        constraint_options["id"] = id_provider.allocate_id(
            sanitize_id(f"location-{resource_id}")
        )
    constraint_options[
        (
            "rsc-pattern"
            if resource_id_type == const.RESOURCE_ID_TYPE_REGEXP
            else "rsc"
        )
    ] = resource_id
    for name, value in constraint_options.items():
        if value != "":
            constraint_el.attrib[name] = value

    # create a rule element
    _add_rule_to_constraint(
        constraint_el, id_provider, cib_schema_version, rule_tree, rule_options
    )

    return constraint_el


def _add_rule_to_constraint(
    constraint_el: _Element,
    id_provider: IdProvider,
    cib_schema_version: Version,
    rule_tree: rule.RuleRoot,
    rule_options: Mapping[str, str],
) -> _Element:
    """
    Add a rule to an existing constraint, remove its simple node-score settings

    constraint_el -- location constraint to be modified
    id_provider -- elements' ids generator
    cib_schema_version -- current CIB schema version
    rule_tree -- parsed rule - specifies when the constraint is active
    rule_options -- additional options for the rule
    """
    rule_options = dict(rule_options)  # make a modifiable copy

    # add the rule to CIB
    rule_el = rule.rule_to_cib(
        constraint_el,
        id_provider,
        cib_schema_version,
        rule_tree,
        rule_options.get("id"),
    )

    # set rule attributes
    if not rule_options.get("score") and not rule_options.get(
        "score-attribute"
    ):
        rule_options["score"] = "INFINITY"
    if rule_options.get("role"):
        rule_options["role"] = get_role_value_for_cib(
            const.PcmkRoleType(rule_options["role"]),
            cib_schema_version >= const.PCMK_NEW_ROLES_CIB_VERSION,
        )
    for name, value in rule_options.items():
        if value != "":
            rule_el.attrib[name] = value

    # If the rule is being added to a location constraint which didn't have
    # rules previously, we need to remove the constraint's node and score
    # attributes to keep CIB valid. The rule basically replaces previous
    # node-score preference.
    for name in ("node", "score"):
        if name in constraint_el.attrib:
            del constraint_el.attrib[name]

    return rule_el



from typing import Mapping

from pcs.common import (
    const,
    reports,
)
from pcs.lib import validate
from pcs.lib.cib.constraint import location
from pcs.lib.cib.tools import (
    ElementNotFound,
    IdProvider,
    get_constraints,
    get_element_by_id,
    get_pacemaker_version_by_which_cib_was_validated,
)
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError


def create_plain_with_rule(
    env: LibraryEnvironment,
    resource_id_type: const.ResourceIdType,
    resource_id: str,
    rule: str,
    rule_options: Mapping[str, str],
    constraint_options: Mapping[str, str],
    force_flags: reports.types.ForceFlags = (),
) -> None:
    """
    Create a location constraint with a rule for a resource

    env --
    resource_id_type -- specifies type of resource_id - plain or regexp
    resource_id -- resource or tag ID or resource ID pattern (regexp)
    rule -- constraint's rule - specifies when the constraint is active
    rule_options -- additional options for the rule
    constraint_options -- additional options for the constraint
    force_flags -- list of flags codes
    """
    # pylint: disable=too-many-locals
    # Pacemaker 3 changed CIB schema for rules. We no longer support the
    # old schema, so we require CIB to be upgraded to the new one.
    cib = env.get_cib(minimal_version=const.PCMK_RULES_PCMK3_SYNTAX_CIB_VERSION)
    id_provider = IdProvider(cib)
    constraint_section = get_constraints(cib)

    # validation
    constrained_el = None
    allowed_id_types = [
        const.RESOURCE_ID_TYPE_PLAIN,
        const.RESOURCE_ID_TYPE_REGEXP,
    ]
    if resource_id_type not in allowed_id_types:
        env.report_processor.report(
            reports.ReportItem.error(
                reports.messages.InvalidIdType(
                    resource_id_type, sorted(allowed_id_types)
                )
            )
        )
    elif resource_id_type == const.RESOURCE_ID_TYPE_PLAIN:
        try:
            constrained_el = get_element_by_id(cib, resource_id)
        except ElementNotFound:
            env.report_processor.report(
                reports.ReportItem.error(
                    reports.messages.IdNotFound(resource_id, [])
                )
            )

    rule_options_pairs = validate.values_to_pairs(
        rule_options,
        validate.option_value_normalization(
            {"role": lambda value: value.capitalize()}
        ),
    )
    validator = location.ValidateCreatePlainWithRule(
        id_provider,
        rule,
        rule_options_pairs,
        constraint_options,
        constrained_el,
    )
    env.report_processor.report_list(validator.validate(force_flags))

    if env.report_processor.has_errors:
        raise LibraryError()

    # modify CIB
    new_constraint = location.create_plain_with_rule(
        constraint_section,
        id_provider,
        get_pacemaker_version_by_which_cib_was_validated(cib),
        resource_id_type,
        resource_id,
        validator.get_parsed_rule(),
        validate.pairs_to_values(rule_options_pairs),
        constraint_options,
    )

    # Check whether the created constraint is a duplicate of an existing one
    env.report_processor.report_list(
        location.DuplicatesCheckerLocationRulePlain().check(
            constraint_section, new_constraint, force_flags
        )
    )
    if env.report_processor.has_errors:
        raise LibraryError()

    # push CIB
    env.push_cib()



FIXTURE_UTILIZATION_WARNING = (
    "Warning: Utilization attributes configuration has no effect until cluster "
    "property option 'placement-strategy' is set to one of the values: "
    "'balanced', 'minimal', 'utilization'\n"
)



import json
from typing import (
    Any,
    Optional,
)

from pcs import (
    resource,
    utils,
)
from pcs.cli.common import parse_args
from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import (
    Argv,
    InputModifiers,
    KeyValueParser,
    ensure_unique_args,
)
from pcs.cli.fencing_topology import target_type_map_cli_to_lib
from pcs.cli.reports.output import (
    print_to_stderr,
)
from pcs.cli.resource.output import resource_agent_metadata_to_text
from pcs.cli.resource.parse_args import (
    parse_primitive as parse_primitive_resource,
)
from pcs.cli.stonith.common import check_is_stonith
from pcs.cli.stonith.levels.output import stonith_level_config_to_text
from pcs.common import reports
from pcs.common.fencing_topology import TARGET_TYPE_ATTRIBUTE
from pcs.common.pacemaker.resource.list import (
    get_all_resources_ids,
    get_stonith_resources_ids,
)
from pcs.common.resource_agent.dto import ResourceAgentNameDto
from pcs.common.str_tools import (
    format_list,
    format_plural,
    indent,
)


def stonith_status_cmd(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    resource.resource_status(lib, argv[:], modifiers, stonith=True)
    if not argv:
        _print_stonith_levels(lib)


def _print_stonith_levels(lib: Any) -> None:
    lines = stonith_level_config_to_text(lib.fencing_topology.get_config_dto())
    if lines:
        print("\nFencing Levels:")
        print("\n".join(indent(lines)))


def stonith_list_available(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --nodesc - do not show description of the agents
    """
    modifiers.ensure_only_supported("--nodesc")
    if len(argv) > 1:
        raise CmdLineInputError()

    search = argv[0] if argv else None
    agent_list = lib.stonith_agent.list_agents(
        describe=not modifiers.get("--nodesc"),
        search=search,
    )

    if not agent_list:
        if search:
            utils.err("No stonith agents matching the filter.")
        utils.err(
            "No stonith agents available. Do you have fence agents installed?"
        )

    for agent_info in agent_list:
        name = agent_info["type"]
        shortdesc = agent_info["shortdesc"]
        if shortdesc:
            print(
                "{0} - {1}".format(
                    name,
                    # pylint: disable=protected-access
                    resource._format_desc(  # noqa: SLF001
                        len(name + " - "), shortdesc.replace("\n", " ")
                    ),
                )
            )
        else:
            print(name)


def stonith_list_options(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --full - show advanced options
    """
    modifiers.ensure_only_supported("--full")
    if len(argv) != 1:
        raise CmdLineInputError()
    agent_name = ResourceAgentNameDto("stonith", None, argv[0])
    print(
        "\n".join(
            resource_agent_metadata_to_text(
                lib.resource_agent.get_agent_metadata(agent_name),
                lib.resource_agent.get_agent_default_operations(
                    agent_name
                ).operations,
                verbose=modifiers.is_specified("--full"),
            )
        )
    )


def stonith_create(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --force - allow not existing agent, invalid operations or invalid
        instance attributes
      * --disabled - created resource will be disabled
      * --no-default-ops - do not add default operations
      * --agent-validation - use agent self validation of instance attributes
      * --wait
      * -f - CIB file
    """
    modifiers.ensure_only_supported(
        "--force",
        "--disabled",
        "--no-default-ops",
        "--agent-validation",
        "--wait",
        "-f",
        hint_syntax_changed="0.12",
    )

    if len(argv) < 2:
        raise CmdLineInputError()

    stonith_id = argv[0]
    stonith_type = argv[1]

    parts = parse_primitive_resource(argv[2:])

    settings = dict(
        allow_absent_agent=modifiers.get("--force"),
        allow_invalid_operation=modifiers.get("--force"),
        allow_invalid_instance_attributes=modifiers.get("--force"),
        ensure_disabled=modifiers.get("--disabled"),
        use_default_operations=not modifiers.get("--no-default-ops"),
        wait=modifiers.get("--wait"),
        enable_agent_self_validation=modifiers.get("--agent-validation"),
    )

    lib.stonith.create(
        stonith_id,
        stonith_type,
        parts.operations,
        parts.meta_attrs,
        parts.instance_attrs,
        **settings,
    )


def _stonith_level_parse_node(arg):
    """
    Commandline options: no options
    """
    target_type_candidate, target_value_candidate = parse_args.parse_typed_arg(
        arg, target_type_map_cli_to_lib.keys(), "node"
    )
    target_type = target_type_map_cli_to_lib[target_type_candidate]
    if target_type == TARGET_TYPE_ATTRIBUTE:
        target_value = parse_args.split_option(target_value_candidate)
    else:
        target_value = target_value_candidate
    return target_type, target_value


def _stonith_level_parse_target_and_stonith(argv: Argv):
    target_type, target_value, devices = None, None, None
    allowed_keywords = {"target", "stonith"}
    missing_target_value, missing_stonith_value = False, False
    groups = parse_args.group_by_keywords(argv, allowed_keywords)
    if groups.has_keyword("target"):
        if len(groups.get_args_flat("target")) > 1:
            raise CmdLineInputError("At most one target can be specified")
        if groups.get_args_flat("target"):
            target_type, target_value = _stonith_level_parse_node(
                groups.get_args_flat("target")[0]
            )
        else:
            missing_target_value = True
    if groups.has_keyword("stonith"):
        if groups.get_args_flat("stonith"):
            devices = groups.get_args_flat("stonith")
        else:
            missing_stonith_value = True

    if missing_target_value and missing_stonith_value:
        raise CmdLineInputError("Missing value after 'target' and 'stonith'")
    if missing_target_value:
        raise CmdLineInputError("Missing value after 'target'")
    if missing_stonith_value:
        raise CmdLineInputError("Missing value after 'stonith'")

    return target_type, target_value, devices


def stonith_level_add_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
      * --force - allow not existing stonith device, allow not existing node
        (target)
    """
    modifiers.ensure_only_supported("-f", "--force")
    if len(argv) < 3:
        raise CmdLineInputError()

    level = argv[0]

    # KeyValueParser supports only key value arguments, so we will only filter
    # out those by looking for "=" to also include mistyped or missing keys
    kvpairs = [arg for arg in argv[2:] if "=" in arg]
    for kvpair in kvpairs:
        argv.remove(kvpair)
    parser = KeyValueParser(kvpairs)
    parser.check_allowed_keys(["id"])
    level_id = parser.get_unique().get("id")
    target_type, target_value = _stonith_level_parse_node(argv[1])
    stonith_devices = argv[2:]
    check_is_stonith(lib, stonith_devices)

    lib.fencing_topology.add_level(
        level,
        target_type,
        target_value,
        stonith_devices,
        level_id=level_id,
        force_device=modifiers.get("--force"),
        force_node=modifiers.get("--force"),
    )


def stonith_level_clear_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f")

    if not argv:
        lib.fencing_topology.remove_all_levels()
        return

    (target_type, target_value, devices) = (
        _stonith_level_parse_target_and_stonith(argv)
    )
    if devices is not None and target_value is not None:
        raise CmdLineInputError(
            "Only one of 'target' and 'stonith' can be used"
        )
    lib.fencing_topology.remove_levels_by_params(
        None,
        target_type,
        target_value,
        devices,
    )


def stonith_level_remove_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f")
    if not argv or len(argv) < 1:
        raise CmdLineInputError()

    level = argv[0]
    target_type, target_value, devices = (
        _stonith_level_parse_target_and_stonith(argv[1:])
    )

    lib.fencing_topology.remove_levels_by_params(
        level,
        target_type,
        target_value,
        devices,
    )


def stonith_level_verify_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f")
    if argv:
        raise CmdLineInputError()
    # raises LibraryError in case of problems, else we don't want to do anything
    lib.fencing_topology.verify()


def stonith_fence(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --off - use off action of fence agent
    """
    del lib
    modifiers.ensure_only_supported("--off")
    if len(argv) != 1:
        utils.err("must specify one (and only one) node to fence")

    node = argv.pop(0)
    if modifiers.get("--off"):
        args = ["stonith_admin", "-F", node]
    else:
        args = ["stonith_admin", "-B", node]
    output, retval = utils.run(args)

    if retval != 0:
        utils.err("unable to fence '%s'\n" % node + output)
    else:
        print_to_stderr(f"Node: {node} fenced")


def stonith_confirm(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --force - required for confirming that fencing happened - DEPRECATED
      * --yes - required for confirming that fencing happened
    """
    del lib
    modifiers.ensure_only_supported("--force", "--yes")
    if len(argv) != 1:
        utils.err("must specify one (and only one) node to confirm fenced")

    node = argv.pop(0)
    if not utils.get_continue_confirmation(
        f"If node '{node}' is not powered off or it does have access to shared "
        "resources, data corruption and/or cluster failure may occur",
        bool(modifiers.get("--yes")),
        bool(modifiers.get("--force")),
    ):
        return
    args = ["stonith_admin", "-C", node]
    output, retval = utils.run(args)

    if retval != 0:
        utils.err("unable to confirm fencing of node '%s'\n" % node + output)
    else:
        print("Node: %s confirmed fenced" % node)


# This is used only by pcsd, will be removed in new architecture
def get_fence_agent_info(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if len(argv) != 1:
        utils.err("One parameter expected")

    agent_name = argv[0]
    if not agent_name.startswith("stonith:"):
        utils.err("Invalid fence agent name")
    print(
        json.dumps(
            lib.stonith_agent.describe_agent(agent_name[len("stonith:") :])
        )
    )


def sbd_watchdog_list(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if argv:
        raise CmdLineInputError()

    available_watchdogs = lib.sbd.get_local_available_watchdogs()

    if available_watchdogs:
        key_label = {
            "identity": "Identity",
            "driver": "Driver",
        }
        lines = []
        for watchdog in sorted(available_watchdogs.keys()):
            lines += [watchdog]
            for key, label in key_label.items():
                value = available_watchdogs[watchdog].get(key)
                if value:
                    lines += indent([f"{label}: {value}"])
        print("Available watchdog(s):\n" + "\n".join(indent(lines)))
    else:
        print_to_stderr("No available watchdog")


def sbd_watchdog_list_json(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if argv:
        raise CmdLineInputError()
    print(json.dumps(lib.sbd.get_local_available_watchdogs()))


def sbd_watchdog_test(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --force - required for testing the watchdog - DEPRECATED
      * --yes - required for testing the watchdog
    """
    modifiers.ensure_only_supported("--force", "--yes")
    if len(argv) > 1:
        raise CmdLineInputError()
    if not utils.get_continue_confirmation(
        "This operation is expected to force-reboot this system without "
        "following any shutdown procedures",
        bool(modifiers.get("--yes")),
        bool(modifiers.get("--force")),
    ):
        return
    watchdog = None
    if len(argv) == 1:
        watchdog = argv[0]
    lib.sbd.test_local_watchdog(watchdog)


def sbd_enable(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --request-timeout - HTTP request timeout
      * --force - allow unknown SBD options
      * --skip-offline - skip offline cluster nodes
      * --no-watchdog-validation - do not validate watchdog
    """
    modifiers.ensure_only_supported(
        "--request-timeout",
        "--force",
        "--skip-offline",
        "--no-watchdog-validation",
    )
    parser = KeyValueParser(argv, repeatable=("device", "watchdog"))
    repeatable_options = parser.get_repeatable()
    default_watchdog, watchdog_dict = _sbd_parse_watchdogs(
        repeatable_options.get("watchdog", [])
    )
    default_device_list, node_device_dict = _sbd_parse_node_specific_options(
        repeatable_options.get("device", [])
    )

    lib.sbd.enable_sbd(
        default_watchdog,
        watchdog_dict,
        parser.get_unique(),
        default_device_list=(
            default_device_list if default_device_list else None
        ),
        node_device_dict=node_device_dict if node_device_dict else None,
        allow_unknown_opts=modifiers.get("--force"),
        ignore_offline_nodes=modifiers.get("--skip-offline"),
        no_watchdog_validation=modifiers.get("--no-watchdog-validation"),
    )


def _sbd_parse_node_specific_options(
    arg_list: Argv,
) -> tuple[list[str], dict[str, list[str]]]:
    """
    Commandline options: no options
    """
    default_option_list = []
    node_specific_option_dict: dict[str, list[str]] = {}

    for arg in arg_list:
        if "@" in arg:
            option, node_name = arg.rsplit("@", 1)
            if node_name in node_specific_option_dict:
                node_specific_option_dict[node_name].append(option)
            else:
                node_specific_option_dict[node_name] = [option]
        else:
            default_option_list.append(arg)

    return default_option_list, node_specific_option_dict


def _sbd_parse_watchdogs(
    watchdog_list: list[str],
) -> tuple[Optional[str], dict[str, str]]:
    """
    Commandline options: no options
    """
    (
        default_watchdog_list,
        node_specific_watchdog_dict,
    ) = _sbd_parse_node_specific_options(watchdog_list)
    if not default_watchdog_list:
        default_watchdog = None
    elif len(default_watchdog_list) == 1:
        default_watchdog = default_watchdog_list[0]
    else:
        raise CmdLineInputError("Multiple watchdog definitions.")

    watchdog_dict = {}
    for node, node_watchdog_list in node_specific_watchdog_dict.items():
        if len(node_watchdog_list) > 1:
            raise CmdLineInputError(
                f"Multiple watchdog definitions for node '{node}'"
            )
        watchdog_dict[node] = node_watchdog_list[0]

    return default_watchdog, watchdog_dict


def sbd_disable(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --request-timeout - HTTP request timeout
      * --skip-offline - skip offline cluster nodes
      * --force - override validation errors
    """
    modifiers.ensure_only_supported(
        "--request-timeout", "--skip-offline", "--force"
    )
    if argv:
        raise CmdLineInputError()

    force_flags = set()
    if modifiers.is_specified("--force"):
        force_flags.add(reports.codes.FORCE)
    if modifiers.is_specified("--skip-offline"):
        force_flags.add(reports.codes.SKIP_OFFLINE_NODES)

    lib.sbd.disable_sbd(modifiers.get("--skip-offline"), force_flags)


def sbd_status(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --request-timeout - HTTP request timeout
      * --full - display SBD device header
    """
    modifiers.ensure_only_supported("--request-timeout", "--full")

    def _bool_to_str(val: Optional[bool]) -> str:
        if val is None:
            return "N/A"
        return "YES" if val else " NO"

    if argv:
        raise CmdLineInputError()

    status_list = lib.sbd.get_cluster_sbd_status()
    if not status_list:
        utils.err("Unable to get SBD status from any node.")

    print("SBD STATUS")
    print("<node name>: <installed> | <enabled> | <running>")
    for node_status in status_list:
        status = node_status["status"]
        print(
            "{node}: {installed} | {enabled} | {running}".format(
                node=node_status["node"],
                installed=_bool_to_str(status.get("installed")),
                enabled=_bool_to_str(status.get("enabled")),
                running=_bool_to_str(status.get("running")),
            )
        )
    device_list = lib.sbd.get_local_devices_info(modifiers.get("--full"))
    for device in device_list:
        print()
        print(f"Messages list on device '{device['device']}':")
        print("<unknown>" if device["list"] is None else device["list"])
        if modifiers.get("--full"):
            print()
            print(f"SBD header on device '{device['device']}':")
            print("<unknown>" if device["dump"] is None else device["dump"])


def _print_per_node_option(config_list, config_option):
    """
    Commandline options: no options
    """
    unknown_value = "<unknown>"
    for config in config_list:
        value = unknown_value
        if config["config"] is not None:
            value = config["config"].get(config_option, unknown_value)
        print(f"  {config['node']}: {value}")


def sbd_config(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --request-timeout - HTTP request timeout
    """
    modifiers.ensure_only_supported("--request-timeout")
    if argv:
        raise CmdLineInputError()

    config_list = lib.sbd.get_cluster_sbd_config()

    if not config_list:
        utils.err("No config obtained.")

    config = config_list[0]["config"]

    filtered_options = [
        "SBD_WATCHDOG_DEV",
        "SBD_OPTS",
        "SBD_PACEMAKER",
        "SBD_DEVICE",
    ]
    with_device = False
    for key, val in config.items():
        if key == "SBD_DEVICE":
            with_device = True
        if key in filtered_options:
            continue
        print(f"{key}={val}")

    print()
    print("Watchdogs:")
    _print_per_node_option(config_list, "SBD_WATCHDOG_DEV")

    if with_device:
        print()
        print("Devices:")
        _print_per_node_option(config_list, "SBD_DEVICE")


def local_sbd_config(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if argv:
        raise CmdLineInputError()
    print(json.dumps(lib.sbd.get_local_sbd_config()))


def sbd_setup_block_device(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --force - required for wiping specified storage devices - DEPRECATED
      * --yes - required for wiping specified storage devices
    """
    modifiers.ensure_only_supported("--force", "--yes")
    parser = KeyValueParser(argv, repeatable=("device",))
    repeatable_options = parser.get_repeatable()
    device_list = repeatable_options.get("device", [])
    if not device_list:
        raise CmdLineInputError("No device defined")

    if not utils.get_continue_confirmation(
        f"All current content on device(s) {format_list(device_list)} will be "
        "overwritten",
        bool(modifiers.get("--yes")),
        bool(modifiers.get("--force")),
    ):
        return

    lib.sbd.initialize_block_devices(device_list, parser.get_unique())


def sbd_message(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if len(argv) != 3:
        raise CmdLineInputError()

    device, node, message = argv
    lib.sbd.set_message(device, node, message)


def stonith_history_show_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if len(argv) > 1:
        raise CmdLineInputError()

    node = argv[0] if argv else None
    print(lib.stonith.history_get_text(node))


def stonith_history_cleanup_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if len(argv) > 1:
        raise CmdLineInputError()

    node = argv[0] if argv else None
    print_to_stderr(lib.stonith.history_cleanup(node))


def stonith_history_update_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options: no options
    """
    modifiers.ensure_only_supported()
    if argv:
        raise CmdLineInputError()

    print_to_stderr(lib.stonith.history_update())


def stonith_update_scsi_devices(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --request-timeout - timeout for HTTP requests
      * --skip-offline - skip unreachable nodes
    """
    modifiers.ensure_only_supported("--request-timeout", "--skip-offline")
    force_flags = []
    if modifiers.get("--skip-offline"):
        force_flags.append(reports.codes.SKIP_OFFLINE_NODES)

    if len(argv) < 2:
        raise CmdLineInputError()
    stonith_id = argv[0]
    parsed_args = parse_args.group_by_keywords(
        argv[1:], ["set", "add", "remove", "delete"]
    )
    parsed_args.ensure_unique_keywords()
    cmd_exception = CmdLineInputError(
        show_both_usage_and_message=True,
        hint=(
            "You must specify either list of set devices or at least one device"
            " for add or delete/remove devices"
        ),
    )
    if parsed_args.has_keyword("set") and (
        parsed_args.has_keyword("add")
        or parsed_args.has_keyword("remove")
        or parsed_args.has_keyword("delete")
    ):
        raise cmd_exception
    if parsed_args.has_keyword("set"):
        if not parsed_args.get_args_flat("set"):
            raise cmd_exception
        lib.stonith.update_scsi_devices(
            stonith_id,
            parsed_args.get_args_flat("set"),
            force_flags=force_flags,
        )
    else:
        for key in ("add", "remove", "delete"):
            if parsed_args.has_empty_keyword(key):
                raise cmd_exception
        lib.stonith.update_scsi_devices_add_remove(
            stonith_id,
            parsed_args.get_args_flat("add"),
            (
                parsed_args.get_args_flat("delete")
                + parsed_args.get_args_flat("remove")
            ),
            force_flags=force_flags,
        )


def delete_cmd(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
      * --force - don't stop a resource before its deletion
    """
    modifiers.ensure_only_supported("-f", "--force")
    if not argv:
        raise CmdLineInputError()
    ensure_unique_args(argv)

    resources_to_remove = set(argv)
    resources_dto = lib.resource.get_configured_resources()
    missing_ids = resources_to_remove - get_all_resources_ids(resources_dto)
    if missing_ids:
        raise CmdLineInputError(
            "Unable to find stonith {resource}: {id_list}".format(
                resource=format_plural(missing_ids, "resource"),
                id_list=format_list(missing_ids),
            )
        )

    non_stonith_ids = resources_to_remove - get_stonith_resources_ids(
        resources_dto
    )
    if non_stonith_ids:
        raise CmdLineInputError(
            (
                "This command cannot remove {resource}: {id_list}. Use 'pcs "
                "resource remove' instead."
            ).format(
                resource=format_plural(non_stonith_ids, "resource"),
                id_list=format_list(non_stonith_ids),
            )
        )

    force_flags = set()
    if modifiers.is_specified("--force"):
        force_flags.add(reports.codes.FORCE)

    lib.cib.remove_elements(resources_to_remove, force_flags)


def enable_cmd(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --wait
      * -f - CIB file
    """
    modifiers.ensure_only_supported("--wait", "-f")
    if not argv:
        raise CmdLineInputError(
            "You must specify stonith resource(s) to enable"
        )
    resources = argv
    check_is_stonith(lib, resources, "pcs resource enable")
    lib.resource.enable(resources, modifiers.get("--wait"))


def disable_cmd(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
      * --force
      * --wait
    """
    if not argv:
        raise CmdLineInputError(
            "You must specify stonith resource(s) to disable"
        )
    check_is_stonith(lib, argv, "pcs resource disable")
    modifiers.ensure_only_supported(
        "-f", "--force", "--wait", hint_syntax_changed="0.12"
    )
    resource.resource_disable_common(lib, argv, modifiers)


def update_cmd(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
      * --wait
      * --force - allow invalid options, do not fail if not possible to get
        agent metadata, allow not suitable command
    """
    if not argv:
        raise CmdLineInputError()
    check_is_stonith(lib, [argv[0]], "pcs resource update")
    resource.resource_update(argv, modifiers)


def op_add_cmd(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
      * --force - allow unknown options
    """
    if not argv:
        raise CmdLineInputError()
    check_is_stonith(lib, [argv[0]], "pcs resource op add")
    resource.resource_op_add(argv, modifiers)


def op_delete_cmd(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f")
    if not argv:
        raise CmdLineInputError()
    check_is_stonith(lib, [argv[0]], "pcs resource op delete")
    resource.resource_operation_remove(argv[0], argv[1:])



from typing import (
    Any,
    Optional,
)

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import (
    FUTURE_OPTION,
    Argv,
    InputModifiers,
    KeyValueParser,
)
from pcs.cli.reports.output import (
    deprecation_warning,
    process_library_reports,
    warn,
)
from pcs.cli.resource.parse_args import (
    parse_primitive as parse_primitive_resource,
)
from pcs.common import reports
from pcs.lib.errors import LibraryError

from .report_processor import NodeRemoveRemoteReportProcessor


def _node_add_remote_separate_name_and_addr(
    arg_list: Argv,
) -> tuple[str, Optional[str], list[str]]:
    """
    Commandline options: no options
    """
    node_name = arg_list[0]
    if len(arg_list) == 1:
        node_addr = None
        rest_args = []
    elif "=" in arg_list[1] or arg_list[1] in ["op", "meta"]:
        node_addr = None
        rest_args = arg_list[1:]
    else:
        node_addr = arg_list[1]
        rest_args = arg_list[2:]
    return node_name, node_addr, rest_args


def node_add_remote(
    lib: Any, arg_list: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --wait
      * --force - allow incomplete distribution of files, allow pcmk remote
        service to fail
      * --skip-offline - skip offline nodes
      * --request-timeout - HTTP request timeout
      * --no-default-ops - do not use default operations
      For tests:
      * --corosync_conf
      * -f
    """
    modifiers.ensure_only_supported(
        "--wait",
        "--force",
        "--skip-offline",
        "--request-timeout",
        "--corosync_conf",
        "-f",
        "--no-default-ops",
    )
    if not arg_list:
        raise CmdLineInputError()

    node_name, node_addr, rest_args = _node_add_remote_separate_name_and_addr(
        arg_list
    )

    parts = parse_primitive_resource(rest_args)
    force = modifiers.get("--force")

    lib.remote_node.node_add_remote(
        node_name,
        node_addr,
        parts.operations,
        parts.meta_attrs,
        parts.instance_attrs,
        skip_offline_nodes=modifiers.get("--skip-offline"),
        allow_incomplete_distribution=force,
        allow_pacemaker_remote_service_fail=force,
        allow_invalid_operation=force,
        allow_invalid_instance_attributes=force,
        use_default_operations=not modifiers.get("--no-default-ops"),
        wait=modifiers.get("--wait"),
    )


def node_remove_remote(
    lib: Any, arg_list: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --force - allow multiple nodes removal, allow pcmk remote service
        to fail, (deprecated) don't stop a resource before its deletion
      * --skip-offline - skip offline nodes
      * --request-timeout - HTTP request timeout
      * --no-stop - don't stop resources before deletion
      * --future - specifying '--force' does not skip resource stopping
      For tests:
      * --corosync_conf
      * -f
    """

    def _process_reports(
        report_list: reports.ReportItemList,
        force_flags: reports.types.ForceFlags,
    ) -> reports.ReportItemList:
        filtered_reports = []
        for report in report_list:
            if (
                report.message.code
                == reports.codes.CANNOT_REMOVE_RESOURCES_NOT_STOPPED
            ):
                continue
            if (
                report.severity.level == reports.ReportItemSeverity.ERROR
                and report.severity.force_code in force_flags
            ):
                report.severity = reports.ReportItemSeverity.warning()
            filtered_reports.append(report)
        return filtered_reports

    modifiers.ensure_only_supported(
        "--corosync_conf",
        "-f",
        "--force",
        FUTURE_OPTION,
        "--no-stop",
        "--request-timeout",
        "--skip-offline",
    )
    modifiers.ensure_not_mutually_exclusive("-f", "--no-stop")

    if len(arg_list) != 1:
        raise CmdLineInputError()
    node_identifier = arg_list[0]

    force_flags = []
    if modifiers.get("--force"):
        force_flags.append(reports.codes.FORCE)
    if modifiers.get("--skip-offline"):
        force_flags.append(reports.codes.SKIP_OFFLINE_NODES)

    if modifiers.is_specified("-f"):
        warn(
            "Resources are not going to be stopped before deletion because the "
            "command does not run on a live cluster"
        )
        lib.remote_node.node_remove_remote(node_identifier, force_flags)
        return

    dont_stop_me_now = modifiers.is_specified("--no-stop")
    if (
        not modifiers.is_specified(FUTURE_OPTION)
        and modifiers.is_specified("--force")
        and not dont_stop_me_now
    ):
        # deprecated after pcs-0.12.0
        deprecation_warning(
            "Using '--force' to skip resource stopping is deprecated and will "
            "be removed in a future release. Specify '--future' to switch to "
            "the future behavior and use '--no-stop' to skip resource stopping."
        )
        dont_stop_me_now = True

    if dont_stop_me_now:
        lib.remote_node.node_remove_remote(node_identifier, force_flags)
        return

    original_report_processor = lib.env.report_processor
    # We cannot use the ReportProcessorInMemory, since we would catch all
    # reports and print them after the command finishes:
    # `remote_node.node_remove_remote` takes a long time (communication with
    # other nodes) => the user would not see any output until the command is
    # finished, possibly making them think the command is stuck and not doing
    # anything.
    temporary_report_processor = NodeRemoveRemoteReportProcessor(
        modifiers.is_specified("--debug")
    )
    lib.env.report_processor = temporary_report_processor

    try:
        temp_force_flags = [
            flag for flag in force_flags if flag != reports.codes.FORCE
        ]
        lib.remote_node.node_remove_remote(node_identifier, temp_force_flags)

        if (
            not temporary_report_processor.already_reported_to_console
            and temporary_report_processor.reports
        ):
            process_library_reports(
                temporary_report_processor.reports,
                include_debug=modifiers.is_specified("--debug"),
            )
        return
    except LibraryError as e:
        # We already printed the errors that caused the LibraryError, so the
        # error is really an error
        if temporary_report_processor.already_reported_to_console:
            raise e

        filtered_reports = _process_reports(
            temporary_report_processor.reports, force_flags
        )

        if reports.has_errors(filtered_reports) or e.output or e.args:
            if filtered_reports:
                process_library_reports(
                    filtered_reports,
                    include_debug=modifiers.is_specified("--debug"),
                    exit_on_error=False,
                )
            raise e

    lib.env.report_processor = original_report_processor

    resource_ids = lib.remote_node.get_resource_ids(node_identifier)
    lib.resource.stop(resource_ids, force_flags)
    lib.cluster.wait_for_pcmk_idle(None)

    lib.remote_node.node_remove_remote(node_identifier, force_flags)


def node_add_guest(lib: Any, arg_list: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --wait
      * --force - allow incomplete distribution of files, allow pcmk remote
        service to fail
      * --skip-offline - skip offline nodes
      * --request-timeout - HTTP request timeout
      For tests:
      * --corosync_conf
      * -f
    """
    modifiers.ensure_only_supported(
        "--wait",
        "--force",
        "--skip-offline",
        "--request-timeout",
        "--corosync_conf",
        "-f",
    )
    if len(arg_list) < 2:
        raise CmdLineInputError()

    node_name = arg_list[0]
    resource_id = arg_list[1]
    meta_options = KeyValueParser(arg_list[2:]).get_unique()

    lib.remote_node.node_add_guest(
        node_name,
        resource_id,
        meta_options,
        skip_offline_nodes=modifiers.get("--skip-offline"),
        allow_incomplete_distribution=modifiers.get("--force"),
        allow_pacemaker_remote_service_fail=modifiers.get("--force"),
        wait=modifiers.get("--wait"),
    )


def node_remove_guest(
    lib: Any, arg_list: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * --wait
      * --force - allow multiple nodes removal, allow pcmk remote service to
        fail
      * --skip-offline - skip offline nodes
      * --request-timeout - HTTP request timeout
      For tests:
      * --corosync_conf
      * -f
    """
    modifiers.ensure_only_supported(
        "--wait",
        "--force",
        "--skip-offline",
        "--request-timeout",
        "--corosync_conf",
        "-f",
    )
    if len(arg_list) != 1:
        raise CmdLineInputError()

    lib.remote_node.node_remove_guest(
        arg_list[0],
        skip_offline_nodes=modifiers.get("--skip-offline"),
        allow_remove_multiple_nodes=modifiers.get("--force"),
        allow_pacemaker_remote_service_fail=modifiers.get("--force"),
        wait=modifiers.get("--wait"),
    )


def node_clear(lib: Any, arg_list: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --force - allow to clear a cluster node
    """
    modifiers.ensure_only_supported("--force")
    if len(arg_list) != 1:
        raise CmdLineInputError()

    lib.cluster.node_clear(
        arg_list[0], allow_clear_cluster_node=modifiers.get("--force")
    )


def cluster_rename(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --force
      * --skip-offline - skip offline nodes
    """
    modifiers.ensure_only_supported("--force", "--skip-offline")
    if len(argv) != 1:
        raise CmdLineInputError()
    force_flags = []
    if modifiers.get("--force"):
        force_flags.append(reports.codes.FORCE)
    if modifiers.get("--skip-offline"):
        force_flags.append(reports.codes.SKIP_OFFLINE_NODES)

    lib.cluster.rename(argv[0], force_flags)



import json
from textwrap import dedent
from unittest import (
    TestCase,
    mock,
)

from pcs.cli.cluster_property import command as cluster_property
from pcs.cli.common.errors import CmdLineInputError
from pcs.common.interface import dto
from pcs.common.pacemaker.cluster_property import ClusterPropertyMetadataDto
from pcs.common.pacemaker.nvset import (
    CibNvpairDto,
    CibNvsetDto,
    ListCibNvsetDto,
)
from pcs.common.reports import codes as report_codes
from pcs.common.resource_agent.dto import ResourceAgentParameterDto

from pcs_test.tools.misc import dict_to_modifiers

FIXTURE_PROPERTY_METADATA = ClusterPropertyMetadataDto(
    properties_metadata=[
        ResourceAgentParameterDto(
            name="property_name",
            shortdesc="Duplicate property",
            longdesc=None,
            type="string",
            default="duplicate_default",
            enum_values=None,
            required=False,
            advanced=False,
            deprecated=False,
            deprecated_by=[],
            deprecated_desc=None,
            unique_group=None,
            reloadable=False,
        ),
        ResourceAgentParameterDto(
            name="property_name",
            shortdesc=None,
            longdesc=None,
            type="string",
            default="default",
            enum_values=None,
            required=False,
            advanced=False,
            deprecated=False,
            deprecated_by=[],
            deprecated_desc=None,
            unique_group=None,
            reloadable=False,
        ),
        ResourceAgentParameterDto(
            name="property_advanced",
            shortdesc=None,
            longdesc=None,
            type="string",
            default="default",
            enum_values=None,
            required=False,
            advanced=True,
            deprecated=False,
            deprecated_by=[],
            deprecated_desc=None,
            unique_group=None,
            reloadable=False,
        ),
    ],
    readonly_properties=["readonly"],
)

FIXTURE_PROPERTIES_CONFIG = ListCibNvsetDto(
    nvsets=[
        CibNvsetDto(
            id="cib-bootstrap-options",
            options={},
            rule=None,
            nvpairs=[
                CibNvpairDto(id="id1", name="property_name", value="value1")
            ],
        )
    ]
)


class TestSetProperty(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["cluster_property"])
        self.cluster_property = mock.Mock(spec_set=["set_properties"])
        self.lib.cluster_property = self.cluster_property

    def _call_cmd(self, argv, modifiers=None):
        cluster_property.set_property(
            self.lib, argv, dict_to_modifiers(modifiers or {})
        )

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([])
        self.assertIsNone(cm.exception.message)
        self.cluster_property.set_properties.assert_not_called()

    def test_empty_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["", ""])
        self.assertEqual(cm.exception.message, "missing value of '' option")
        self.cluster_property.set_properties.assert_not_called()

    def test_option_missing_value(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["a=1", "b"])
        self.assertEqual(cm.exception.message, "missing value of 'b' option")
        self.cluster_property.set_properties.assert_not_called()

    def test_value_missing_option_name(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["=1", "b=2"])
        self.assertEqual(cm.exception.message, "missing key in '=1' option")
        self.cluster_property.set_properties.assert_not_called()

    def test_option_multiple_values(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["a=1", "a=2", "b=2", "b="])
        self.assertEqual(
            cm.exception.message,
            "duplicate option 'a' with different values '1' and '2'",
        )
        self.cluster_property.set_properties.assert_not_called()

    def test_multiple_args(self):
        self._call_cmd(["a=1", "b=2", "c="])
        self.cluster_property.set_properties.assert_called_once_with(
            {"a": "1", "b": "2", "c": ""}, set()
        )

    def test_multiple_args_with_force(self):
        self._call_cmd(["a=1", "b=2", "c="], {"force": True})
        self.cluster_property.set_properties.assert_called_once_with(
            {"a": "1", "b": "2", "c": ""}, {report_codes.FORCE}
        )

    def test_unsupported_modifier(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([], {"all": True})
        self.assertEqual(
            cm.exception.message,
            "Specified option '--all' is not supported in this command",
        )
        self.cluster_property.assert_not_called()


class TestUnsetProperty(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["cluster_property"])
        self.cluster_property = mock.Mock(spec_set=["set_properties"])
        self.lib.cluster_property = self.cluster_property

    def _call_cmd(self, argv, modifiers=None):
        cluster_property.unset_property(
            self.lib, argv, dict_to_modifiers(modifiers or {})
        )

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([])
        self.assertIsNone(cm.exception.message)
        self.cluster_property.set_properties.assert_not_called()

    def test_args(self):
        self._call_cmd(["a=1", "=b", ""])
        self.cluster_property.set_properties.assert_called_once_with(
            {"a=1": "", "=b": "", "": ""}, set()
        )

    def test_args_with_force(self):
        self._call_cmd(["a=1", "=b", ""], {"force": True})
        self.cluster_property.set_properties.assert_called_once_with(
            {"a=1": "", "=b": "", "": ""}, {report_codes.FORCE}
        )

    def test_unsupported_modifier(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([], {"defaults": True})
        self.assertEqual(
            cm.exception.message,
            "Specified option '--defaults' is not supported in this command",
        )
        self.cluster_property.assert_not_called()

    def test_duplicate_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["a", "a", "b", "b", "c"])
        self.assertEqual(
            cm.exception.message,
            "duplicate arguments: 'a', 'b'",
        )
        self.cluster_property.assert_not_called()


@mock.patch("pcs.cli.cluster_property.command.print")
class TestConfig(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["cluster_property"])
        self.cluster_property = mock.MagicMock(
            spec_set=["get_properties", "get_properties_metadata"]
        )
        self.cluster_property.get_properties_metadata.return_value = (
            FIXTURE_PROPERTY_METADATA
        )
        self.cluster_property.get_properties.return_value = (
            FIXTURE_PROPERTIES_CONFIG
        )
        self.lib.cluster_property = self.cluster_property

    def _call_cmd(self, argv, modifiers=None):
        cluster_property.config(
            self.lib, argv, dict_to_modifiers(modifiers or {})
        )

    def test_default_output_format(self, mock_print):
        self._call_cmd([])
        self.cluster_property.get_properties.assert_called_once()
        self.cluster_property.get_properties_metadata.assert_not_called()
        mock_print.assert_called_once_with(
            dedent(
                """\
                Cluster Properties: cib-bootstrap-options
                  property_name=value1"""
            )
        )

    def test_specified_properties(self, mock_print):
        self._call_cmd(["property_name", "property_advanced", "property"])
        self.cluster_property.get_properties.assert_called_once()
        self.cluster_property.get_properties_metadata.assert_called_once()
        mock_print.assert_called_once_with(
            dedent(
                """\
                Cluster Properties: cib-bootstrap-options
                  property_advanced=default (default)
                  property_name=value1"""
            )
        )

    def test_output_format_cmd(self, mock_print):
        self._call_cmd([], {"output-format": "cmd"})
        self.cluster_property.get_properties.assert_called_once()
        self.cluster_property.get_properties_metadata.assert_called_once()
        mock_print.assert_called_once_with(
            "pcs property set --force -- \\\n  property_name=value1"
        )

    def test_all_option(self, mock_print):
        self._call_cmd([], {"all": True})
        self.cluster_property.get_properties.assert_called_once()
        self.cluster_property.get_properties_metadata.assert_called_once()
        mock_print.assert_called_once_with(
            dedent(
                """\
                Cluster Properties: cib-bootstrap-options
                  property_advanced=default (default)
                  property_name=value1"""
            )
        )

    @mock.patch("pcs.cli.cluster_property.command.deprecation_warning")
    def test_defaults_option(self, mock_warn, mock_print):
        self._call_cmd([], {"defaults": True})
        self.cluster_property.get_properties.assert_not_called()
        self.cluster_property.get_properties_metadata.assert_called_once()
        mock_print.assert_called_once_with(
            dedent(
                """\
                property_advanced=default
                property_name=default"""
            )
        )
        mock_warn.assert_called_once_with(
            "Option --defaults is deprecated and will be "
            "removed. Please use command 'pcs property defaults' instead."
        )

    def test_output_format_json(self, mock_print):
        self._call_cmd([], {"output-format": "json"})
        self.cluster_property.get_properties.assert_called_once()
        self.cluster_property.get_properties_metadata.assert_not_called()
        mock_print.assert_called_once_with(
            json.dumps(dto.to_dict(FIXTURE_PROPERTIES_CONFIG))
        )

    def test_unsupported_option(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([], {"force": True})
        self.assertEqual(
            cm.exception.message,
            ("Specified option '--force' is not supported in this command"),
        )
        self.cluster_property.get_properties.assert_not_called()
        self.cluster_property.get_properties_metadata.assert_not_called()
        mock_print.assert_not_called()

    def test_args_and_option(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["arg1"], {"all": True})
        self.assertEqual(
            cm.exception.message,
            (
                "cannot specify properties when using '--all', '--defaults', "
                "'--output-format'"
            ),
        )
        self.cluster_property.get_properties.assert_not_called()
        self.cluster_property.get_properties_metadata.assert_not_called()
        mock_print.assert_not_called()

    def test_unsupported_output_format(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([], {"output-format": "unsupported"})
        self.assertEqual(
            cm.exception.message,
            (
                "Unknown value 'unsupported' for '--output-format' option. "
                "Supported values are: 'cmd', 'json', 'text'"
            ),
        )
        self.cluster_property.get_properties.assert_not_called()
        self.cluster_property.get_properties_metadata.assert_not_called()
        mock_print.assert_not_called()

    def test_all_modifiers(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(
                [], {"all": True, "defaults": True, "output-format": "text"}
            )
        self.assertEqual(
            cm.exception.message,
            (
                "Only one of '--all', '--defaults', '--output-format' can be "
                "used"
            ),
        )
        self.cluster_property.get_properties.assert_not_called()
        self.cluster_property.get_properties_metadata.assert_not_called()
        mock_print.assert_not_called()

    def test_all_and_defaults(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([], {"all": True, "defaults": True})
        self.assertEqual(
            cm.exception.message,
            ("Only one of '--all', '--defaults' can be used"),
        )
        self.cluster_property.get_properties.assert_not_called()
        self.cluster_property.get_properties_metadata.assert_not_called()
        mock_print.assert_not_called()

    def test_all_and_output_format(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([], {"all": True, "output-format": True})
        self.assertEqual(
            cm.exception.message,
            ("Only one of '--all', '--output-format' can be used"),
        )
        self.cluster_property.get_properties.assert_not_called()
        self.cluster_property.get_properties_metadata.assert_not_called()
        mock_print.assert_not_called()

    def test_defaults_and_output_format(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([], {"defaults": True, "output-format": True})
        self.assertEqual(
            cm.exception.message,
            ("Only one of '--defaults', '--output-format' can be used"),
        )
        self.cluster_property.get_properties.assert_not_called()
        self.cluster_property.get_properties_metadata.assert_not_called()
        mock_print.assert_not_called()


@mock.patch("pcs.cli.cluster_property.command.print")
class TestDefaults(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["cluster_property"])
        self.cluster_property = mock.MagicMock(
            spec_set=["get_properties_metadata"]
        )
        self.cluster_property.get_properties_metadata.return_value = (
            FIXTURE_PROPERTY_METADATA
        )
        self.lib.cluster_property = self.cluster_property

    def _call_cmd(self, argv, modifiers=None):
        cluster_property.defaults(
            self.lib, argv, dict_to_modifiers(modifiers or {})
        )

    def test_no_args(self, mock_print):
        self._call_cmd([])
        self.cluster_property.get_properties_metadata.assert_called_once()
        mock_print.assert_called_once_with("property_name=default")

    def test_full_option(self, mock_print):
        self._call_cmd([], {"full": True})
        self.cluster_property.get_properties_metadata.assert_called_once()
        mock_print.assert_called_once_with(
            "property_advanced=default\nproperty_name=default"
        )

    def test_filter_properties(self, mock_print):
        self._call_cmd(["property_name", "property_advanced"])
        self.cluster_property.get_properties_metadata.assert_called_once()
        mock_print.assert_called_once_with(
            "property_advanced=default\nproperty_name=default"
        )

    def test_properties_without_defaults(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["arg1", "arg2"])
        self.assertEqual(
            cm.exception.message,
            "No default value for properties: 'arg1', 'arg2'",
        )
        self.cluster_property.get_properties_metadata.assert_called_once()
        mock_print.assert_not_called()

    def test_unsupported_option(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([], modifiers={"force": True, "full": True})
        self.assertEqual(
            cm.exception.message,
            "Specified option '--force' is not supported in this command",
        )
        self.cluster_property.get_properties_metadata.assert_not_called()
        mock_print.assert_not_called()

    def test_args_and_full(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["arg1", "arg2"], modifiers={"full": True})
        self.assertEqual(
            cm.exception.message,
            "cannot specify properties when using '--full'",
        )
        self.cluster_property.get_properties_metadata.assert_not_called()
        mock_print.assert_not_called()


@mock.patch("pcs.cli.cluster_property.command.print")
class TestDescribe(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["cluster_property"])
        self.cluster_property = mock.MagicMock(
            spec_set=["get_properties_metadata"]
        )
        self.cluster_property.get_properties_metadata.return_value = (
            FIXTURE_PROPERTY_METADATA
        )
        self.lib.cluster_property = self.cluster_property
        self.text_output = dedent(
            """\
            property_name
              Description: No description available
              Type: string
              Default: default"""
        )
        self.text_output_advanced = dedent(
            """\
            property_advanced (advanced use only)
              Description: No description available
              Type: string
              Default: default
            property_name
              Description: No description available
              Type: string
              Default: default"""
        )

    def _call_cmd(self, argv, modifiers=None):
        cluster_property.describe(
            self.lib, argv, dict_to_modifiers(modifiers or {})
        )

    def test_no_args(self, mock_print):
        self._call_cmd([])
        self.cluster_property.get_properties_metadata.assert_called_once()
        mock_print.assert_called_once_with(self.text_output)

    def test_full_option(self, mock_print):
        self._call_cmd([], {"full": True})
        self.cluster_property.get_properties_metadata.assert_called_once()
        mock_print.assert_called_once_with(self.text_output_advanced)

    def test_filter_properties(self, mock_print):
        self._call_cmd(["property_name", "property_advanced"])
        self.cluster_property.get_properties_metadata.assert_called_once()
        mock_print.assert_called_once_with(self.text_output_advanced)

    def test_properties_without_metadata(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["arg1", "arg2"])
        self.assertEqual(
            cm.exception.message,
            "No description for properties: 'arg1', 'arg2'",
        )
        self.cluster_property.get_properties_metadata.assert_called_once()
        mock_print.assert_not_called()

    def test_supported_output_format_json(self, mock_print):
        self._call_cmd([], modifiers={"output-format": "json"})
        self.cluster_property.get_properties_metadata.assert_called_once_with()
        mock_print.assert_called_once_with(
            json.dumps(dto.to_dict(FIXTURE_PROPERTY_METADATA))
        )

    def test_unsupported_output_format(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([], modifiers={"output-format": "cmd"})
        self.assertEqual(
            cm.exception.message,
            "Unknown value 'cmd' for '--output-format' option. Supported values"
            " are: 'json', 'text'",
        )
        self.cluster_property.get_properties_metadata.assert_not_called()
        mock_print.assert_not_called()

    def test_unsupported_option(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([], modifiers={"force": True, "full": True})
        self.assertEqual(
            cm.exception.message,
            "Specified option '--force' is not supported in this command",
        )
        self.cluster_property.get_properties_metadata.assert_not_called()
        mock_print.assert_not_called()

    def test_args_and_full(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["arg1", "arg2"], modifiers={"full": True})
        self.assertEqual(
            cm.exception.message,
            "cannot specify properties when using '--full'",
        )
        self.cluster_property.get_properties_metadata.assert_not_called()
        mock_print.assert_not_called()

    def test_output_format_json_and_full(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(
                [], modifiers={"output-format": "json", "full": True}
            )
        self.assertEqual(
            cm.exception.message,
            "property filtering is not supported with --output-format=json",
        )
        self.cluster_property.get_properties_metadata.assert_not_called()
        mock_print.assert_not_called()


@mock.patch("pcs.cli.cluster_property.command.print")
class TestPrintClusterPropertiesDefinitionLegacy(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["cluster_property"])
        self.cluster_property = mock.MagicMock(
            spec_set=["get_cluster_properties_definition_legacy"]
        )
        self.cluster_property.get_cluster_properties_definition_legacy.return_value = {}
        self.lib.cluster_property = self.cluster_property

    def _call_cmd(self, argv, modifiers=None):
        cluster_property.print_cluster_properties_definition_legacy(
            self.lib, argv, dict_to_modifiers(modifiers or {})
        )

    def test_args(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["args"])
        self.assertIsNone(cm.exception.message)
        self.cluster_property.get_cluster_properties_definition_legacy.assert_not_called()
        mock_print.assert_not_called()

    def test_options(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([], modifiers={"force": True, "full": True})
        self.assertEqual(
            cm.exception.message,
            (
                "Specified options '--force', '--full' are not supported in "
                "this command"
            ),
        )
        self.cluster_property.get_cluster_properties_definition_legacy.assert_not_called()
        mock_print.assert_not_called()

    def test_no_args(self, mock_print):
        self._call_cmd([])
        self.cluster_property.get_cluster_properties_definition_legacy.assert_called_once()
        mock_print.assert_called_once_with("{}")



from collections import defaultdict
from dataclasses import dataclass
from enum import Enum, auto
from typing import Final, Iterable, Literal, Optional, Sequence, Union, cast

from pcs.common.const import (
    PCMK_ROLE_STOPPED,
    PCMK_STATUS_ROLE_DEMOTING,
    PCMK_STATUS_ROLE_MIGRATING,
    PCMK_STATUS_ROLE_PROMOTED,
    PCMK_STATUS_ROLE_PROMOTING,
    PCMK_STATUS_ROLE_STARTED,
    PCMK_STATUS_ROLE_STARTING,
    PCMK_STATUS_ROLE_STOPPED,
    PCMK_STATUS_ROLE_STOPPING,
    PCMK_STATUS_ROLE_UNPROMOTED,
    PCMK_STATUS_ROLES_PENDING,
)
from pcs.common.status_dto import (
    AnyResourceStatusDto,
    BundleReplicaStatusDto,
    BundleStatusDto,
    CloneStatusDto,
    GroupStatusDto,
    PrimitiveStatusDto,
    ResourcesStatusDto,
)


class ResourceType(Enum):
    PRIMITIVE = "primitive"
    GROUP = "group"
    CLONE = "clone"
    BUNDLE = "bundle"


# used to check if any value other than None is present
# in optional attributes of status dto
class NotNoneValue:
    pass


NOT_NONE: Final = NotNoneValue()


StateValueType = Union[str, bool, NotNoneValue, set[str]]

AttributeTuple = tuple[str, StateValueType]


class ResourceState(Enum):
    """
    possible values for checking the state of the resource
    """

    STARTED = ("role", PCMK_STATUS_ROLE_STARTED)
    STOPPED = ("role", PCMK_STATUS_ROLE_STOPPED)
    PROMOTED = ("role", PCMK_STATUS_ROLE_PROMOTED)
    UNPROMOTED = ("role", PCMK_STATUS_ROLE_UNPROMOTED)
    STARTING = ("role", PCMK_STATUS_ROLE_STARTING)
    STOPPING = ("role", PCMK_STATUS_ROLE_STOPPING)
    MIGRATING = ("role", PCMK_STATUS_ROLE_MIGRATING)
    PROMOTING = ("role", PCMK_STATUS_ROLE_PROMOTING)
    DEMOTING = ("role", PCMK_STATUS_ROLE_DEMOTING)
    MONITORING = ("pending", "Monitoring")
    DISABLED = ("disabled", True)
    ENABLED = ("disabled", False)
    MANAGED = ("managed", True)
    UNMANAGED = ("managed", False)
    MAINTENANCE = ("maintenance", True)
    FAILED = ("failed", True)
    ACTIVE = ("active", True)
    ORPHANED = ("orphaned", True)
    BLOCKED = ("blocked", True)
    FAILURE_IGNORED = ("failure_ignored", True)
    PENDING = [
        ("role", set(PCMK_STATUS_ROLES_PENDING)),
        ("pending", "Monitoring"),
    ]
    LOCKED_TO = ("locked_to", NOT_NONE)


ResourceStateExactCheck = Literal[ResourceState.LOCKED_TO]

EXACT_CHECK_STATES = (ResourceState.LOCKED_TO,)


class MoreChildrenQuantifierType(Enum):
    ALL = auto()
    ANY = auto()
    NONE = auto()


@dataclass(frozen=True)
class PrimitiveInstances:
    instances: Sequence[PrimitiveStatusDto]


@dataclass(frozen=True)
class GroupInstances:
    instances: Sequence[GroupStatusDto]


CheckedResourceType = Union[
    PrimitiveInstances,
    GroupInstances,
    CloneStatusDto,
    BundleStatusDto,
]

_TYPE_MAP = {
    PrimitiveStatusDto: ResourceType.PRIMITIVE,
    GroupStatusDto: ResourceType.GROUP,
    CloneStatusDto: ResourceType.CLONE,
    BundleStatusDto: ResourceType.BUNDLE,
}


_UNIQUE_RESOURCES_TYPES = (CloneStatusDto, BundleStatusDto)
_UNIQUE_RESOURCES_CONST = tuple(
    _TYPE_MAP[res_type] for res_type in _UNIQUE_RESOURCES_TYPES
)

_PROMOTABLE_RESOURCES_TYPES = (CloneStatusDto,)
_PROMOTABLE_RESOURCES_CONST = tuple(
    _TYPE_MAP[res_type] for res_type in _PROMOTABLE_RESOURCES_TYPES
)


def none(iterable: Iterable[object]) -> bool:
    return not any(iterable)


_MORE_CHILDREN_QUANTIFIER_MAP = {
    MoreChildrenQuantifierType.ALL: all,
    MoreChildrenQuantifierType.ANY: any,
    MoreChildrenQuantifierType.NONE: none,
}


class QueryException(Exception):
    pass


class MembersQuantifierUnsupportedException(QueryException):
    pass


class InstancesQuantifierUnsupportedException(QueryException):
    pass


class ResourceException(Exception):
    def __init__(self, resource_id: str, instance_id: Optional[str]):
        self.resource_id = resource_id
        self.instance_id = instance_id


class ResourceNonExistentException(ResourceException):
    pass


class ResourceNotInGroupException(ResourceException):
    pass


class ResourceUnexpectedTypeException(ResourceException):
    def __init__(
        self,
        resource_id: str,
        instance_id: Optional[str],
        resource_type: ResourceType,
        expected_types: list[ResourceType],
    ):
        super().__init__(resource_id, instance_id)
        self.resource_type = resource_type
        self.expected_types = expected_types


def can_be_unique(resource_type: ResourceType) -> bool:
    return resource_type in _UNIQUE_RESOURCES_CONST


def can_be_promotable(resource_type: ResourceType) -> bool:
    return resource_type in _PROMOTABLE_RESOURCES_CONST


class ResourcesStatusFacade:
    def __init__(self, resources: Sequence[AnyResourceStatusDto]):
        self._resources = resources
        self._resource_map: dict[str, list[AnyResourceStatusDto]] = defaultdict(
            list
        )
        self._child_parent_map: dict[str, str] = {}
        self._bundle_member_node_map: dict[int, list[str]] = {}
        self._bundle_defined = False

        for resource in resources:
            self._resource_map[resource.resource_id].append(resource)
            if isinstance(resource, GroupStatusDto):
                self.__add_group_children_to_maps(resource)
            elif isinstance(resource, CloneStatusDto):
                for instance in resource.instances:
                    self._child_parent_map[instance.resource_id] = (
                        resource.resource_id
                    )
                    self._resource_map[instance.resource_id].append(instance)
                    if isinstance(instance, GroupStatusDto):
                        self.__add_group_children_to_maps(instance)
            elif isinstance(resource, BundleStatusDto):
                self._bundle_defined = True
                for replica in resource.replicas:
                    member = replica.member
                    if member is not None:
                        self._child_parent_map[member.resource_id] = (
                            resource.resource_id
                        )
                        self._resource_map[member.resource_id].append(member)
                        self._bundle_member_node_map[id(member)] = (
                            replica.container.node_names
                        )

    @classmethod
    def from_resources_status_dto(
        cls, resources_status_dto: ResourcesStatusDto
    ) -> "ResourcesStatusFacade":
        """
        Create ResourcesStatusFacade from ResourcesStatusDto

        resources_status_dto -- dto with status of the resources in cluster
        """
        return cls(resources_status_dto.resources)

    def get_resource_one_instance(
        self, resource_id: str, instance_id: Optional[str]
    ) -> Optional[AnyResourceStatusDto]:
        """
        Get one instance of resource with given id. Get instance that
        appears first in the status xml if instance_id is not specified. Return
        None if resource with given id does not exist.

        resource_id -- id of the resource
        instance_id -- id describing unique instance of cloned or bundled
            resource
        """
        resource_list = self._resource_map.get(resource_id)
        if resource_list is None:
            return None

        if instance_id is None:
            return resource_list[0]

        for resource in resource_list:
            if (
                hasattr(resource, "instance_id")
                and resource.instance_id == instance_id
            ):
                return resource
        return None

    def get_resource_all_instances(
        self, resource_id: str
    ) -> Optional[list[AnyResourceStatusDto]]:
        """
        Get a list of all the instances of resource with the given id. Return
        None if resource with the given id does not exist.

        resource_id -- id of the resource
        """
        return self._resource_map.get(resource_id)

    def exists(self, resource_id: str, instance_id: Optional[str]) -> bool:
        """
        Check if resource with the given id exists in the cluster

        resource_id -- id of the resource
        instance_id -- id describing unique instance of cloned or bundled
            resource
        """
        return (
            self.get_resource_one_instance(resource_id, instance_id) is not None
        )

    def get_type(
        self, resource_id: str, instance_id: Optional[str]
    ) -> ResourceType:
        """
        Return the type of the resource

        resource_id -- id of the resource
        instance_id -- id describing unique instance of cloned or bundled
            resource
        """
        resource = self.get_resource_one_instance(resource_id, instance_id)
        if resource is None:
            raise ResourceNonExistentException(resource_id, instance_id)

        return _TYPE_MAP[type(resource)]

    def is_stonith(self, resource_id: str, instance_id: Optional[str]) -> bool:
        """
        Check if the resource with the given id is a stonith resource

        resource_id -- id of the resource
        instance_id -- id describing unique instance of cloned or bundled
            resource
        """
        resource = self.get_resource_one_instance(resource_id, instance_id)
        if resource is None:
            raise ResourceNonExistentException(resource_id, instance_id)

        return isinstance(
            resource, PrimitiveStatusDto
        ) and resource.resource_agent.startswith("stonith:")

    def is_promotable(
        self, resource_id: str, instance_id: Optional[str]
    ) -> bool:
        """
        Check if the resource with the given id is promotable. Usable only
        of clone resources.

        resource_id -- id of the resource
        instance_id -- id describing unique instance of cloned or bundled
            resource
        """
        resource = self.get_resource_one_instance(resource_id, instance_id)
        if resource is None:
            raise ResourceNonExistentException(resource_id, instance_id)

        if not isinstance(resource, _PROMOTABLE_RESOURCES_TYPES):
            raise ResourceUnexpectedTypeException(
                resource_id,
                instance_id,
                _TYPE_MAP[type(resource)],
                list(_PROMOTABLE_RESOURCES_CONST),
            )

        return resource.multi_state

    def is_unique(self, resource_id: str, instance_id: Optional[str]) -> bool:
        """
        Check if the resource with the given id is globally unique. Usable only
        on clone and bundle resources.

        resource_id -- id of the resource
        instance_id -- id describing unique instance of cloned or bundled
            resource
        """

        resource = self.get_resource_one_instance(resource_id, instance_id)
        if resource is None:
            raise ResourceNonExistentException(resource_id, instance_id)

        if not isinstance(resource, _UNIQUE_RESOURCES_TYPES):
            raise ResourceUnexpectedTypeException(
                resource_id,
                instance_id,
                _TYPE_MAP[type(resource)],
                list(_UNIQUE_RESOURCES_CONST),
            )

        return resource.unique

    def _get_instances_for_state_check(
        self, resource_id: str, instance_id: Optional[str]
    ) -> CheckedResourceType:
        if not self.exists(resource_id, instance_id):
            raise ResourceNonExistentException(resource_id, instance_id)

        is_in_clone = (
            self.get_type(resource_id, instance_id)
            in (ResourceType.PRIMITIVE, ResourceType.GROUP)
            and self.get_parent_clone_id(resource_id, instance_id) is not None
        )
        is_clone = self.get_type(resource_id, instance_id) == ResourceType.CLONE

        if self._bundle_defined and (is_in_clone or is_clone):
            # This is due to pacemaker providing broken data in this case.
            # See issue: https://projects.clusterlabs.org/T722
            raise NotImplementedError(
                "Queries on clone state when bundle resources are present "
                "in the cluster are unsupported"
            )

        resource = self.get_resource_one_instance(resource_id, instance_id)
        if isinstance(resource, (CloneStatusDto, BundleStatusDto)):
            return resource

        if instance_id is None:
            instance_list = cast(
                Optional[list[Union[PrimitiveStatusDto, GroupStatusDto]]],
                self.get_resource_all_instances(resource_id),
            )
        else:
            instance_list = [resource] if resource is not None else None

        if instance_list is None:
            raise ResourceNonExistentException(resource_id, instance_id)

        if is_in_clone:
            instance_list = _filter_clone_orphans(instance_list)

        if not instance_list:
            raise ResourceNonExistentException(resource_id, instance_id)

        if self.get_type(resource_id, instance_id) == ResourceType.PRIMITIVE:
            return PrimitiveInstances(
                cast(list[PrimitiveStatusDto], instance_list)
            )

        return GroupInstances(cast(list[GroupStatusDto], instance_list))

    def can_have_multiple_members(
        self, resource_id: str, instance_id: Optional[str] = None
    ) -> bool:
        """
        Check if the resource with the given id can have multiple inner members.

        resource_id -- id of the resource
        instance_id -- id describing unique instance of cloned or bundled
            resource
        """
        resource_type = self.get_type(resource_id, instance_id)
        return resource_type == ResourceType.GROUP or (
            resource_type == ResourceType.CLONE
            and any(
                self.get_type(member_id, None) == ResourceType.GROUP
                for member_id in self.get_members(resource_id, instance_id)
            )
        )

    def can_have_multiple_instances(
        self, resource_id: str, instance_id: Optional[str] = None
    ) -> bool:
        """
        Check if the resource with the given id can have multiple instances.

        resource_id -- id of the resource
        instance_id -- id describing unique instance of cloned or bundled
            resource
        """
        resource_type = self.get_type(resource_id, instance_id)
        return instance_id is None and (
            resource_type in (ResourceType.CLONE, ResourceType.BUNDLE)
            or self.get_parent_clone_id(resource_id, None) is not None
            or (
                resource_type == ResourceType.PRIMITIVE
                and self.get_parent_bundle_id(resource_id, None) is not None
            )
        )

    def _validate_quantifiers(
        self,
        resource_id: str,
        instance_id: Optional[str],
        members_quantifier: Optional[MoreChildrenQuantifierType],
        instances_quantifier: Optional[MoreChildrenQuantifierType],
    ) -> None:
        if (
            members_quantifier is not None
            and not self.can_have_multiple_members(resource_id, instance_id)
        ):
            raise MembersQuantifierUnsupportedException()
        if (
            instances_quantifier is not None
            and not self.can_have_multiple_instances(resource_id, instance_id)
        ):
            raise InstancesQuantifierUnsupportedException()

    def is_state(
        self,
        resource_id: str,
        instance_id: Optional[str],
        state: ResourceState,
        expected_node_name: Optional[str] = None,
        members_quantifier: Optional[MoreChildrenQuantifierType] = None,
        instances_quantifier: Optional[MoreChildrenQuantifierType] = None,
    ) -> bool:
        """
        Check if the resource with the given id is in expected state.
        With groups, the state is read and evaluated on the group first. If the
        state cannot be determined from only the group, evaluate the state of
        the member resources and return true if the query is true for ALL of
        the members. With clones and bundles, the state is read and evaluated
        on the clone or bundle first. If the state cannot be determined from
        only the clone or bundle, evaluate the state on the instances and
        return true if the query is true for ANY of the instances.

        resource_id -- id of the resource
        instance_id -- id describing unique instance of cloned or bundled
            resource
        state -- expected state of the resource
        expected_node_name -- Check if the resource is in given state on a
            given node. If specified, always check only group members, clone
            instances and bundle replicas even when the attribute is present
            in the group, clone or bundle status dto respectively.
        members_quantifier -- Specify how to treat group members. If specified,
            check only member resources even when the attribute is present in
            the group status dto.
        instances_quantifier -- Specify how to treat instances of clones and
            bundles. If specified, check only instances even when the attribute
            is present in the clone/bundle status dto.
        """
        resource = self._get_instances_for_state_check(resource_id, instance_id)

        self._validate_quantifiers(
            resource_id, instance_id, members_quantifier, instances_quantifier
        )

        if not isinstance(state.value, list):
            checked_state = [state.value]
        else:
            checked_state = state.value

        return any(
            self._check_resources_state_attributes(
                resource,
                checked_attribute,
                real_expected_value,
                expected_node_name,
                members_quantifier,
                instances_quantifier,
            )
            for checked_attribute, real_expected_value in checked_state
        )

    def is_state_exact_value(
        self,
        resource_id: str,
        instance_id: Optional[str],
        state: ResourceStateExactCheck,
        expected_state_value: str,
        expected_node_name: Optional[str] = None,
        members_quantifier: Optional[MoreChildrenQuantifierType] = None,
        instances_quantifier: Optional[MoreChildrenQuantifierType] = None,
    ) -> bool:
        """
        Check if the state attribute of the resource contains the expected
        value. With groups, the state is read and evaluated on the group first.
        If the state cannot be determined from only the group, evaluate the
        state of the member resources and return true if the query is true for
        ALL of the members. With clones and bundles, the state is read and
        evaluated on the clone or bundle first. If the state cannot be
        determined from only the clone or bundle, evaluate the state on the
        instances and return true if the query is true for ANY of the instances.

        resource_id -- id of the resource
        instance_id -- id describing unique instance of cloned or bundled
            resource
        state -- state describing which state attribute will be checked
        expected_state_value -- expected value of the attribute
        expected_node_name -- Check if the resource is in given state on a
            given node. If specified, always check only group members, clone
            instances and bundle replicas even when the attribute is present
            in the group, clone or bundle status dto respectively.
        members_quantifier -- Specify how to treat group members. If specified,
            check only member resources even when the attribute is present in
            the group status dto.
        instances_quantifier -- Specify how to treat instances of clones and
            bundles. If specified, check only instances even when the attribute
            is present in the clone/bundle status dto.
        """
        resource = self._get_instances_for_state_check(resource_id, instance_id)

        self._validate_quantifiers(
            resource_id, instance_id, members_quantifier, instances_quantifier
        )

        if not isinstance(state.value, list):
            checked_state = [state.value]
        else:
            checked_state = state.value

        return any(
            self._check_resources_state_attributes(
                resource,
                checked_attribute,
                expected_state_value,
                expected_node_name,
                members_quantifier,
                instances_quantifier,
            )
            for checked_attribute, _ in checked_state
        )

    def get_parent_group_id(
        self,
        resource_id: str,
        instance_id: Optional[str],
    ) -> Optional[str]:
        """
        Check if the resource is in any group and return group id if the
        resource is in group.

        resource_id -- id of the resource
        instance_id -- id describing unique instance of cloned or bundled
            resource
        """
        resource = self.get_resource_one_instance(resource_id, instance_id)
        if resource is None:
            raise ResourceNonExistentException(resource_id, instance_id)

        if not isinstance(resource, PrimitiveStatusDto):
            raise ResourceUnexpectedTypeException(
                resource_id,
                instance_id,
                _TYPE_MAP[type(resource)],
                [ResourceType.PRIMITIVE],
            )

        if not self._check_parent_type(resource_id, ResourceType.GROUP):
            return None

        return self._child_parent_map[resource_id]

    def get_parent_clone_id(
        self,
        resource_id: str,
        instance_id: Optional[str],
    ) -> Optional[str]:
        """
        Check if the resource is inside any clone and return clone id if the
        resource is in clone. Member of a cloned group is in clone as well.

        resource_id -- id of the resource
        instance_id -- id describing unique instance of cloned or bundled
            resource
        """
        resource = self.get_resource_one_instance(resource_id, instance_id)
        if resource is None:
            raise ResourceNonExistentException(resource_id, instance_id)

        if not isinstance(resource, (PrimitiveStatusDto, GroupStatusDto)):
            raise ResourceUnexpectedTypeException(
                resource_id,
                instance_id,
                _TYPE_MAP[type(resource)],
                [ResourceType.PRIMITIVE, ResourceType.GROUP],
            )

        checked_id = resource_id
        if self._check_parent_type(checked_id, ResourceType.GROUP):
            checked_id = self._child_parent_map[resource_id]

        if not self._check_parent_type(checked_id, ResourceType.CLONE):
            return None

        return self._child_parent_map[checked_id]

    def get_parent_bundle_id(
        self,
        resource_id: str,
        instance_id: Optional[str],
    ) -> Optional[str]:
        """
        Check if the resource is inside any bundle and return bundle id if the
        resource is in bundle.

        resource_id -- id of the resource
        instance_id -- id describing unique instance of cloned or bundled
            resource
        """
        resource = self.get_resource_one_instance(resource_id, instance_id)
        if resource is None:
            raise ResourceNonExistentException(resource_id, instance_id)

        if not isinstance(resource, PrimitiveStatusDto):
            raise ResourceUnexpectedTypeException(
                resource_id,
                instance_id,
                _TYPE_MAP[type(resource)],
                [ResourceType.PRIMITIVE],
            )

        if not self._check_parent_type(resource_id, ResourceType.BUNDLE):
            return None

        return self._child_parent_map[resource_id]

    def get_index_in_group(
        self, resource_id: str, instance_id: Optional[str]
    ) -> int:
        """
        Return the index of the resource in a group. Usable only for primitive
        resources.

        resource_id -- id of the resource
        instance_id -- id describing unique instance of cloned or bundled
            resource
        """
        if not self.exists(resource_id, instance_id):
            raise ResourceNonExistentException(resource_id, instance_id)

        resource_type = self.get_type(resource_id, instance_id)
        if resource_type != ResourceType.PRIMITIVE:
            raise ResourceUnexpectedTypeException(
                resource_id,
                instance_id,
                resource_type,
                [ResourceType.PRIMITIVE],
            )

        parent_id = self._child_parent_map.get(resource_id, None)
        if parent_id is None:
            raise ResourceNotInGroupException(resource_id, instance_id)

        parent_resource = self.get_resource_one_instance(parent_id, None)
        if not isinstance(parent_resource, GroupStatusDto):
            raise ResourceNotInGroupException(resource_id, instance_id)

        return [res.resource_id for res in parent_resource.members].index(
            resource_id
        )

    def get_members(
        self, resource_id: str, instance_id: Optional[str]
    ) -> list[str]:
        """
        Return resource ids of members of a group, clone or bundle

        resource_id -- id of the resource
        instance_id -- id describing unique instance of cloned or bundled
            resource
        """
        resource = self.get_resource_one_instance(resource_id, instance_id)
        if resource is None:
            raise ResourceNonExistentException(resource_id, instance_id)

        if isinstance(resource, GroupStatusDto):
            return [member.resource_id for member in resource.members]

        if isinstance(resource, CloneStatusDto):
            return list(
                {
                    instance.resource_id
                    for instance in resource.instances
                    if not _is_orphaned(instance)
                }
            )

        if isinstance(resource, BundleStatusDto):
            return list(
                {
                    replica.member.resource_id
                    for replica in resource.replicas
                    if replica.member is not None
                }
            )

        raise ResourceUnexpectedTypeException(
            resource_id,
            instance_id,
            ResourceType.PRIMITIVE,
            [ResourceType.GROUP, ResourceType.CLONE, ResourceType.BUNDLE],
        )

    def get_nodes(
        self, resource_id: str, instance_id: Optional[str]
    ) -> list[str]:
        """
        Return nodes on which resource is running. For groups, return nodes
        on which any of the members is running. For clones or bundles return
        nodes on which any of the instances or replicas are running.

        resource_id -- id of the resource
        instance_id -- id describing unique instance of cloned or bundled
            resource
        """
        if instance_id is not None:
            resource = self.get_resource_one_instance(resource_id, instance_id)
            instance_list = [resource] if resource is not None else None
        else:
            instance_list = self.get_resource_all_instances(resource_id)

        if instance_list is None:
            raise ResourceNonExistentException(resource_id, instance_id)

        nodes = []
        for instance in instance_list:
            if isinstance(instance, PrimitiveStatusDto):
                nodes.extend(self._get_primitive_nodes(instance))
            elif isinstance(instance, GroupStatusDto):
                nodes.extend(self._get_group_nodes(instance))
            elif isinstance(instance, CloneStatusDto):
                nodes.extend(self._get_clone_nodes(instance))
            else:
                nodes.extend(self._get_bundle_nodes(instance))

        return sorted(nodes)

    def _get_primitive_nodes(self, primitive: PrimitiveStatusDto) -> list[str]:
        if id(primitive) in self._bundle_member_node_map:
            return self._bundle_member_node_map[id(primitive)]
        return primitive.node_names

    def _get_group_nodes(self, group: GroupStatusDto) -> list[str]:
        node_set = set()
        for member in group.members:
            node_set.update(self._get_primitive_nodes(member))
        return list(node_set)

    def _get_clone_nodes(self, clone: CloneStatusDto) -> list[str]:
        node_list = []
        instance_list = _filter_clone_orphans(clone.instances)
        for instance in instance_list:
            if isinstance(instance, PrimitiveStatusDto):
                node_list.extend(self._get_primitive_nodes(instance))
            elif isinstance(instance, GroupStatusDto):
                node_list.extend(self._get_group_nodes(instance))
        return node_list

    def _get_bundle_nodes(self, bundle: BundleStatusDto) -> list[str]:
        node_list = []
        for replica in bundle.replicas:
            if replica.member is not None:
                node_list.extend(self._get_primitive_nodes(replica.member))
            else:
                node_list.extend(self._get_primitive_nodes(replica.container))
        return node_list

    def _check_parent_type(
        self,
        resource_id: str,
        expected_parent_type: ResourceType,
    ) -> bool:
        parent_id = self._child_parent_map.get(resource_id)
        return (
            parent_id is not None
            and self.get_type(parent_id, None) == expected_parent_type
        )

    def __add_group_children_to_maps(self, group: GroupStatusDto) -> None:
        for child in group.members:
            self._resource_map[child.resource_id].append(child)
            self._child_parent_map[child.resource_id] = group.resource_id

    def _check_resources_state_attributes(
        self,
        checked_resource: CheckedResourceType,
        checked_attribute: str,
        expected_attribute_value: StateValueType,
        expected_node_name: Optional[str] = None,
        members_quantifier: Optional[MoreChildrenQuantifierType] = None,
        instances_quantifier: Optional[MoreChildrenQuantifierType] = None,
    ) -> bool:
        if isinstance(checked_resource, CloneStatusDto):
            return self._clone_state(
                checked_resource,
                checked_attribute,
                expected_attribute_value,
                members_quantifier,
                instances_quantifier,
                expected_node_name,
            )

        if isinstance(checked_resource, BundleStatusDto):
            return self._bundle_state(
                checked_resource,
                checked_attribute,
                expected_attribute_value,
                instances_quantifier,
                expected_node_name,
            )

        more_children_check = (
            instances_quantifier or MoreChildrenQuantifierType.ANY
        )

        if isinstance(checked_resource, PrimitiveInstances):
            return _MORE_CHILDREN_QUANTIFIER_MAP[more_children_check](
                self._primitive_state(
                    resource,
                    checked_attribute,
                    expected_attribute_value,
                    expected_node_name,
                )
                for resource in checked_resource.instances
            )

        if isinstance(checked_resource, GroupInstances):
            return _MORE_CHILDREN_QUANTIFIER_MAP[more_children_check](
                self._group_state(
                    resource,
                    checked_attribute,
                    expected_attribute_value,
                    members_quantifier,
                    expected_node_name,
                )
                for resource in checked_resource.instances
            )

        return False

    def _primitive_state(
        self,
        primitive: PrimitiveStatusDto,
        checked_attribute: str,
        expected_attribute_value: StateValueType,
        expected_node_name: Optional[str],
    ) -> bool:
        if checked_attribute == ResourceState.DISABLED.value[0]:
            result = (
                primitive.target_role == PCMK_ROLE_STOPPED
            ) == expected_attribute_value
        else:
            result = _check_attribute_value(
                primitive, checked_attribute, expected_attribute_value
            )

        if expected_node_name is None:
            return result

        node_names = primitive.node_names
        if id(primitive) in self._bundle_member_node_map:
            node_names = self._bundle_member_node_map[id(primitive)]

        return result and expected_node_name in node_names

    def _group_state(
        self,
        group: GroupStatusDto,
        checked_attribute: str,
        expected_attribute_value: StateValueType,
        members_quantifier: Optional[MoreChildrenQuantifierType],
        expected_node_name: Optional[str],
    ) -> bool:
        if _can_check_non_primitive(
            group,
            checked_attribute,
            members_quantifier,
            expected_node_name is not None,
        ):
            return _check_attribute_value(
                group, checked_attribute, expected_attribute_value
            )

        if not group.members:
            return False

        if members_quantifier is None:
            members_quantifier = MoreChildrenQuantifierType.ALL

        return _MORE_CHILDREN_QUANTIFIER_MAP[members_quantifier](
            self._primitive_state(
                primitive,
                checked_attribute,
                expected_attribute_value,
                expected_node_name,
            )
            for primitive in group.members
        )

    def _clone_state(
        self,
        clone: CloneStatusDto,
        checked_attribute: str,
        expected_attribute_value: StateValueType,
        members_quantifier: Optional[MoreChildrenQuantifierType],
        instances_quantifier: Optional[MoreChildrenQuantifierType],
        expected_node_name: Optional[str],
    ) -> bool:
        if _can_check_non_primitive(
            clone,
            checked_attribute,
            instances_quantifier,
            expected_node_name is not None,
        ):
            return _check_attribute_value(
                clone, checked_attribute, expected_attribute_value
            )

        instance_list = _filter_clone_orphans(clone.instances)

        if not instance_list:
            return False

        if instances_quantifier is None:
            instances_quantifier = MoreChildrenQuantifierType.ANY

        return _MORE_CHILDREN_QUANTIFIER_MAP[instances_quantifier](
            (
                self._primitive_state(
                    instance,
                    checked_attribute,
                    expected_attribute_value,
                    expected_node_name,
                )
                if isinstance(instance, PrimitiveStatusDto)
                else self._group_state(
                    instance,
                    checked_attribute,
                    expected_attribute_value,
                    members_quantifier,
                    expected_node_name,
                )
            )
            for instance in instance_list
        )

    def _bundle_state(
        self,
        bundle: BundleStatusDto,
        checked_attribute: str,
        expected_attribute_value: StateValueType,
        instances_quantifier: Optional[MoreChildrenQuantifierType],
        expected_node_name: Optional[str],
    ) -> bool:
        if _can_check_non_primitive(
            bundle,
            checked_attribute,
            instances_quantifier,
            expected_node_name is not None,
        ):
            return _check_attribute_value(
                bundle, checked_attribute, expected_attribute_value
            )

        if not bundle.replicas:
            return False

        if instances_quantifier is None:
            instances_quantifier = MoreChildrenQuantifierType.ANY

        return _MORE_CHILDREN_QUANTIFIER_MAP[instances_quantifier](
            self._bundle_replica_state(
                replica,
                checked_attribute,
                expected_attribute_value,
                expected_node_name,
            )
            for replica in bundle.replicas
        )

    def _bundle_replica_state(
        self,
        replica: BundleReplicaStatusDto,
        checked_attribute: str,
        expected_attribute_value: StateValueType,
        expected_node_name: Optional[str],
    ) -> bool:
        if replica.member is None:
            return self._primitive_state(
                replica.container,
                checked_attribute,
                expected_attribute_value,
                expected_node_name,
            )

        return self._primitive_state(
            replica.member,
            checked_attribute,
            expected_attribute_value,
            expected_node_name,
        )


def _can_check_non_primitive(
    resource: AnyResourceStatusDto,
    attribute_name: str,
    more_children_check: Optional[MoreChildrenQuantifierType],
    check_nodes: bool,
) -> bool:
    return (
        hasattr(resource, attribute_name)
        and more_children_check is None
        # we have to look at primitives if we want to see on which nodes
        # the resource is in given state
        and not check_nodes
    )


def _is_orphaned(resource: Union[PrimitiveStatusDto, GroupStatusDto]) -> bool:
    if isinstance(resource, PrimitiveStatusDto):
        return resource.orphaned
    return all(child.orphaned for child in resource.members)


def _filter_clone_orphans(
    instance_list: Sequence[Union[PrimitiveStatusDto, GroupStatusDto]],
) -> list[Union[PrimitiveStatusDto, GroupStatusDto]]:
    return [
        instance for instance in instance_list if not _is_orphaned(instance)
    ]


def _check_attribute_value(
    status_dto: AnyResourceStatusDto,
    attribute_name: str,
    expected_value: StateValueType,
) -> bool:
    real_value = getattr(status_dto, attribute_name)

    if isinstance(expected_value, set):
        return real_value in expected_value
    if expected_value == NOT_NONE:
        return real_value is not None

    return real_value == expected_value



from pcs.common import reports
from pcs.common.file import RawFileError
from pcs.common.pcs_cfgsync_dto import SyncConfigsDto
from pcs.lib.env import LibraryEnvironment, LibraryError
from pcs.lib.file.instance import FileInstance
from pcs.lib.file.raw_file import raw_file_error_report
from pcs.lib.pcs_cfgsync.const import SYNCED_CONFIGS


def get_configs(env: LibraryEnvironment, cluster_name: str) -> SyncConfigsDto:
    """
    Get contents of synced configuration files from node

    cluster_name -- expected cluster name. End with an error if the requested
        node is not in the cluster with the expected name.
    """
    current_cluster_name = env.get_corosync_conf().get_cluster_name()
    if current_cluster_name != cluster_name:
        env.report_processor.report(
            reports.ReportItem.error(
                reports.messages.NodeReportsUnexpectedClusterName(cluster_name)
            )
        )
    if env.report_processor.has_errors:
        raise LibraryError()

    configs = {}
    for file_type_code in SYNCED_CONFIGS:
        file_instance = FileInstance.for_common(file_type_code)
        if not file_instance.raw_file.exists():
            # it's not an error if the file does not exist locally, we just
            # wont send it back
            continue
        try:
            configs[file_type_code] = file_instance.read_raw().decode("utf-8")
        except RawFileError as e:
            # in case of error when reading some file, we still might be able
            # to read and send the others without issues
            env.report_processor.report(
                raw_file_error_report(e, is_forced_or_warning=True)
            )

    return SyncConfigsDto(current_cluster_name, configs)



from unittest import (
    TestCase,
    mock,
)

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.constraint_colocation import command as colocation_command
from pcs.common.pacemaker.constraint import (
    CibConstraintColocationAttributesDto,
    CibConstraintColocationDto,
    CibConstraintsDto,
)

from pcs_test.tools.misc import dict_to_modifiers

FIXTURE_COLOCATION_CONSTRAINTS_DTO = CibConstraintsDto(
    colocation=[
        CibConstraintColocationDto(
            resource_id="A",
            with_resource_id="B",
            node_attribute=None,
            resource_role=None,
            with_resource_role=None,
            resource_instance=None,
            with_resource_instance=None,
            attributes=CibConstraintColocationAttributesDto(
                constraint_id="colocation-A-B",
                score=None,
                influence=None,
                lifetime=[],
            ),
        ),
        CibConstraintColocationDto(
            resource_id="B",
            with_resource_id="A",
            node_attribute=None,
            resource_role=None,
            with_resource_role=None,
            resource_instance=None,
            with_resource_instance=None,
            attributes=CibConstraintColocationAttributesDto(
                constraint_id="colocation-B-A",
                score=None,
                influence=None,
                lifetime=[],
            ),
        ),
        CibConstraintColocationDto(
            resource_id="C",
            with_resource_id="D",
            node_attribute=None,
            resource_role=None,
            with_resource_role=None,
            resource_instance=None,
            with_resource_instance=None,
            attributes=CibConstraintColocationAttributesDto(
                constraint_id="colocation-C-D",
                score=None,
                influence=None,
                lifetime=[],
            ),
        ),
    ]
)


class TestRemoveColocationConstraint(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["constraint", "cib"])
        self.cib = mock.Mock(spec_set=["remove_elements"])
        self.constraint = mock.Mock(spec_set=["get_config"])
        self.constraint.get_config.return_value = (
            FIXTURE_COLOCATION_CONSTRAINTS_DTO
        )
        self.lib.cib = self.cib
        self.lib.constraint = self.constraint

    def _call_cmd(self, argv, modifiers=None):
        colocation_command.remove(
            self.lib, argv, dict_to_modifiers(modifiers or {})
        )

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([])
        self.assertIsNone(cm.exception.message)
        self.constraint.get_config.assert_not_called()
        self.cib.remove_elements.assert_not_called()

    def test_not_enough_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["id1"])
        self.assertIsNone(cm.exception.message)
        self.constraint.get_config.assert_not_called()
        self.cib.remove_elements.assert_not_called()

    def test_too_many_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["id1", "id2", "id3"])
        self.assertIsNone(cm.exception.message)
        self.constraint.get_config.assert_not_called()
        self.cib.remove_elements.assert_not_called()

    def test_unsupported_modifier(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["id1", "id2"], {"force": True})
        self.assertEqual(
            cm.exception.message,
            "Specified option '--force' is not supported in this command",
        )
        self.constraint.get_config.assert_not_called()
        self.cib.remove_elements.assert_not_called()

    def test_colocation_constraint_not_found(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["A", "D"])
        self.assertEqual(
            cm.exception.message,
            (
                "Unable to find colocation constraint with source resource "
                "'A' and target resource 'D'"
            ),
        )
        self.constraint.get_config.assert_called_once_with(evaluate_rules=False)
        self.cib.remove_elements.assert_not_called()

    def test_colocation_constraint_found(self):
        colocation_constraints_ids = ["C", "D"]
        self._call_cmd(colocation_constraints_ids)
        self.constraint.get_config.assert_called_once_with(evaluate_rules=False)
        self.cib.remove_elements.assert_called_once_with(["colocation-C-D"])

    @mock.patch("pcs.cli.constraint_colocation.command.deprecation_warning")
    def test_remove_colocation_constraint_with_interchanged_ids(
        self, mock_warn
    ):
        colocation_constraints_ids = ["A", "B"]
        self._call_cmd(colocation_constraints_ids)
        self.constraint.get_config.assert_called_once_with(evaluate_rules=False)
        self.cib.remove_elements.assert_called_once_with(
            ["colocation-A-B", "colocation-B-A"]
        )
        mock_warn.assert_called_once_with(
            "Removing colocation constraint with interchanged source resource "
            "id and target resource id. This behavior is deprecated and will "
            "be removed."
        )



import json
from typing import Any

import pcs.lib.pacemaker.live as lib_pacemaker
from pcs import utils
from pcs.cli.cluster_property.output import PropertyConfigurationFacade
from pcs.cli.common.errors import (
    ERR_NODE_LIST_AND_ALL_MUTUALLY_EXCLUSIVE,
    CmdLineInputError,
)
from pcs.cli.common.parse_args import (
    OUTPUT_FORMAT_OPTION,
    Argv,
    InputModifiers,
    KeyValueParser,
)


def node_attribute_cmd(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file (in lib wrapper)
      * --force - no error if attribute to delete doesn't exist
      * --name - specify attribute name for filter
      * --output-format - supported formats: text, cmd, json
    """
    del lib
    modifiers.ensure_only_supported(
        "-f", "--force", "--name", output_format_supported=True
    )
    if len(argv) < 2 or modifiers.is_specified_any(
        ["--name", OUTPUT_FORMAT_OPTION]
    ):
        raise CmdLineInputError()
    attribute_set_cmd(argv.pop(0), argv)


def node_utilization_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Options:
      * -f - CIB file (in lib wrapper)
      * --name - specify attribute name for filter
      * --output-format - supported formats: text, cmd, json
    """
    modifiers.ensure_only_supported(
        "-f", "--name", output_format_supported=True
    )
    if len(argv) < 2 or modifiers.is_specified_any(
        ["--name", OUTPUT_FORMAT_OPTION]
    ):
        raise CmdLineInputError()
    utils.print_warning_if_utilization_attrs_has_no_effect(
        PropertyConfigurationFacade.from_properties_dtos(
            lib.cluster_property.get_properties(),
            lib.cluster_property.get_properties_metadata(),
        )
    )
    set_node_utilization(argv.pop(0), argv)


def node_maintenance_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers, enable: bool
) -> None:
    """
    Options:
      * -f - CIB file (in lib wrapper)
      * --wait - wait for node to change state
      * --all - all cluster nodes
    """
    modifiers.ensure_only_supported("-f", "--wait", "--all")
    if argv and modifiers.get("--all"):
        raise CmdLineInputError(ERR_NODE_LIST_AND_ALL_MUTUALLY_EXCLUSIVE)
    wait = modifiers.get("--wait")
    if modifiers.get("--all"):
        lib.node.maintenance_unmaintenance_all(enable, wait)
    elif argv:
        lib.node.maintenance_unmaintenance_list(enable, argv, wait)
    else:
        # -f cannot be used when editing local node
        lib.node.maintenance_unmaintenance_local(enable, wait)


def node_standby_cmd(
    lib: Any, argv: Argv, modifiers: InputModifiers, enable: bool
) -> None:
    """
    Options:
      * -f - CIB file (in lib wrapper)
      * --wait - wait for node to change state
      * --all - all cluster nodes
    """
    modifiers.ensure_only_supported("-f", "--wait", "--all")
    if argv and modifiers.get("--all"):
        raise CmdLineInputError(ERR_NODE_LIST_AND_ALL_MUTUALLY_EXCLUSIVE)
    wait = modifiers.get("--wait")
    if modifiers.get("--all"):
        lib.node.standby_unstandby_all(enable, wait)
    elif argv:
        lib.node.standby_unstandby_list(enable, argv, wait)
    else:
        # -f cannot be used when editing local node
        lib.node.standby_unstandby_local(enable, wait)


def set_node_utilization(node: str, argv: Argv) -> None:
    """
    Commandline options:
      * -f - CIB file
    """
    nvpair_dict = KeyValueParser(argv).get_unique()
    if not nvpair_dict:
        return
    only_removing = True
    for value in nvpair_dict.values():
        if value != "":
            only_removing = False
            break

    cib = utils.get_cib_dom()
    node_el = utils.dom_get_node(cib, node)
    if node_el is None:
        if utils.usefile:
            utils.err(f"Unable to find a node: {node}")

        for attrs in utils.getNodeAttributesFromPacemaker():
            if attrs.name == node and attrs.type == "remote":
                node_attrs = attrs
                break
        else:
            utils.err(f"Unable to find a node: {node}")

        nodes_section_list = cib.getElementsByTagName("nodes")
        if not nodes_section_list:
            utils.err("Unable to get nodes section of cib")

        if only_removing:
            # Do not create new node if we are only removing values from it.
            return

        dom = nodes_section_list[0].ownerDocument
        node_el = dom.createElement("node")
        node_el.setAttribute("id", node_attrs.id)
        node_el.setAttribute("type", node_attrs.type)
        node_el.setAttribute("uname", node_attrs.name)
        nodes_section_list[0].appendChild(node_el)

    utils.dom_update_utilization(node_el, nvpair_dict, "nodes-")
    utils.replace_cib_configuration(cib)


def node_pacemaker_status(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    """
    Internal pcs-pcsd command
    """
    del lib
    del argv
    del modifiers
    print(json.dumps(lib_pacemaker.get_local_node_status(utils.cmd_runner())))


def attribute_set_cmd(node: str, argv: Argv) -> None:
    """
    Commandline options:
      * -f - CIB file
      * --force - no error if attribute to delete doesn't exist
    """
    for name, value in KeyValueParser(argv).get_unique().items():
        utils.set_node_attribute(name, value, node)
